import sys

from cube.networks.utils import unpack, mask_concat

sys.path.append('')
import os, argparse

os.environ["TOKENIZERS_PARALLELISM"] = "false"
import pytorch_lightning as pl
import torch.nn as nn
import torch.nn.functional as F
import torch
from cube.io_utils.objects import Document, Sentence, Token, Word
from cube.io_utils.encodings import Encodings
from cube.io_utils.config import TokenizerConfig
from cube.networks.utils_tokenizer import TokenCollate
import numpy as np
from cube.networks.modules import ConvNorm, LinearNorm, MLP
from torch.utils.data import DataLoader
import random

from cube.networks.modules import WordGram


class Tokenizer(pl.LightningModule):
    def __init__(self, config: TokenizerConfig, encodings: Encodings, language_codes: [] = None, ext_word_emb=0,
                 max_seq_len=-1):
        super().__init__()
        self._language_codes = language_codes
        self._config = config
        self._max_seq_len = max_seq_len
        if not isinstance(ext_word_emb, list):
            ext_word_emb = [ext_word_emb]
        self._ext_word_emb = ext_word_emb
        conv_layers = []
        cs_inp = config.external_proj_size + config.lang_emb_size + 256 + 16
        NUM_FILTERS = config.cnn_filter
        for _ in range(config.cnn_layers):
            conv_layer = nn.Sequential(
                ConvNorm(cs_inp,
                         NUM_FILTERS,
                         kernel_size=5, stride=1,
                         padding=2,
                         dilation=1, w_init_gain='tanh'),
                nn.BatchNorm1d(NUM_FILTERS))
            conv_layers.append(conv_layer)
            cs_inp = NUM_FILTERS // 2 + config.lang_emb_size
        self._convs = nn.ModuleList(conv_layers)
        self._wg = WordGram(len(encodings.char2int), num_langs=encodings.num_langs)
        self._lang_emb = nn.Embedding(encodings.num_langs + 1, config.lang_emb_size, padding_idx=0)
        self._spa_emb = nn.Embedding(3, 16, padding_idx=0)
        self._output = LinearNorm(NUM_FILTERS // 2 + config.lang_emb_size, 5)

        ext2int = []
        for input_size in self._ext_word_emb:
            module = MLP(input_size, config.external_proj_size)
            ext2int.append(module)
        self._ext_proj = nn.ModuleList(ext2int)

        if self._language_codes:  # only for training
            self._dev_results = {i: [] for i, _ in enumerate(self._language_codes)}

            self._res = {}
            for language_code in self._language_codes:
                self._res[language_code] = {"sent": 0., "token": 0.}
            self._early_stop_meta_val = 0
            self._epoch_results = {}

    def forward(self, batch):
        x_emb = batch['x_input']
        x_spa = batch['x_input_spa']
        x_lang = batch['x_lang']
        x_lang = self._lang_emb(x_lang).unsqueeze(1).repeat(1, x_emb[0].shape[1], 1)
        x_word_char = batch['x_word_char']
        x_word_case = batch['x_word_case']
        x_word_lang = batch['x_word_lang']
        x_word_masks = batch['x_word_masks']
        x_word_len = batch['x_word_len']
        x_sent_len = batch['x_sent_len']
        char_emb_packed = self._wg(x_word_char, x_word_case, x_word_lang, x_word_masks, x_word_len)

        sl = x_sent_len.cpu().numpy()

        x_char_emb = unpack(char_emb_packed, sl, x_emb[0].shape[1], device=self._get_device())

        word_emb_ext = None

        for ii in range(len(x_emb)):
            we = x_emb[ii]
            if word_emb_ext is None:
                word_emb_ext = self._ext_proj[ii](we.float().to(self._get_device()))
            else:
                word_emb_ext = word_emb_ext + self._ext_proj[ii](we)

        word_emb_ext = word_emb_ext / len(x_emb)
        word_emb_ext = torch.tanh(word_emb_ext)
        x_emb = word_emb_ext
        x_spa_emb = self._spa_emb(x_spa)
        x_emb = mask_concat([x_emb, x_char_emb], 0.33, self.training, self._get_device())
        x_emb = torch.cat([x_emb, x_spa_emb], dim=-1)
        x = torch.cat([x_emb, x_lang], dim=-1).permute(0, 2, 1)
        x_lang = x_lang.permute(0, 2, 1)
        half = self._config.cnn_filter // 2
        res = None
        cnt = 0
        for conv in self._convs:
            conv_out = conv(x)
            tmp = torch.tanh(conv_out[:, :half, :]) * torch.sigmoid((conv_out[:, half:, :]))
            if res is None:
                res = tmp
            else:
                res = res + tmp
            x = torch.dropout(tmp, 0.2, self.training)
            cnt += 1
            if cnt != self._config.cnn_layers:
                x = torch.cat([x, x_lang], dim=1)
        x = x + res
        x = torch.cat([x, x_lang], dim=1)
        x = x.permute(0, 2, 1)
        return self._output(x)

    def validation_step(self, batch, batch_idx):
        if batch['x_input'] is None:
            print("Return 0")
            return None
        x_lang = batch['x_lang']
        x_text = batch['x_text']
        y_offset = batch['y_offset'].cpu().numpy()
        y_target = batch['y_output'].cpu().numpy()
        y_len = batch['y_len'].cpu().numpy()
        x_l = x_lang.cpu().numpy()
        y_pred = self.forward(batch)
        y_pred = torch.argmax(y_pred, dim=-1).detach().cpu().numpy()
        for ii in range(len(y_len)):
            ofs = y_offset[ii]
            lang = x_l[ii] - 1
            for jj in range(y_len[ii]):
                self._dev_results[lang].append([x_text[ii][jj], y_target[ii, jj + ofs], y_pred[ii, jj + ofs]])

    def validation_epoch_end(self, outputs) -> None:
        # empty accumulator
        # results = {langid: {'SENT_F': 0, 'TOK_F': 0} for langid in self._id2lang}
        results = {}

        for lang in self._dev_results:
            data = self._dev_results[lang]
            g_sents = []
            p_sents = []
            tok_p = ''
            tok_g = ''
            g_sent = []
            p_sent = []
            for example in data:
                target = example[1]
                pred = example[2]
                text = example[0].replace('▁', '')
                tok_g += text
                tok_p += text
                if target == 2 or target == 3 or target == 4:
                    if tok_g.strip() != '':
                        g_sent.append(tok_g)
                    tok_g = ''
                if target == 4:
                    if len(g_sent) != 0:
                        g_sents.append(g_sent)
                    g_sent = []

                if pred == 2 or pred == 3 or pred == 4:
                    if tok_p.strip() != '':
                        p_sent.append(tok_p)
                    tok_p = ''
                if pred == 4:
                    if len(p_sent) != 0:
                        p_sents.append(p_sent)
                    p_sent = []

            if tok_g.strip() != '':
                g_sent.append(tok_g)
            if len(g_sent) != 0:
                g_sents.append(g_sent)
            if tok_p.strip() != '':
                p_sent.append(tok_p)
            if len(p_sent) != 0:
                p_sents.append(p_sent)

            sent_f, tok_f = _conll_eval(g_sents, p_sents)

            if self._language_codes is not None:
                lang = self._language_codes[lang]
            results[lang] = {}
            results[lang]['sent'] = sent_f
            results[lang]['token'] = tok_f
            self.log('val/SENT/{0}'.format(lang), sent_f)
            self.log('val/TOKEN/{0}'.format(lang), tok_f)

        self._dev_results = {i: [] for i, _ in enumerate(self._language_codes)}
        self._epoch_results = self._compute_early_stop(results)
        self.log('val/early_meta', self._early_stop_meta_val)

    def training_step(self, batch, batch_idx):
        if batch['x_input'] is None:
            print("Return 0")
            return None

        y_target = batch['y_output']
        if self._max_seq_len != -1 and y_target.shape[1] > self._max_seq_len:  # fix for HF
            return None
        y_pred = self.forward(batch)

        loss = F.cross_entropy(y_pred.view(-1, y_pred.shape[2]), y_target.view(-1), ignore_index=0)
        return loss

    def load(self, model_path: str, device: str = 'cpu'):
        self.load_state_dict(torch.load(model_path, map_location='cpu')['state_dict'])
        self.to(device)

    def process(self, raw_text, collate: TokenCollate, batch_size=32, num_workers: int = 4, lang_id: int = 0):
        raw_text = raw_text.replace('\n', ' ').replace('\r', ' ')
        new_text = raw_text.replace('  ', ' ')
        while new_text != raw_text:
            raw_text = new_text
            new_text = raw_text.replace('  ', ' ')

        self.eval()
        from cube.networks.utils_tokenizer import TokenDatasetLive
        dataset = TokenDatasetLive(raw_text, collate.get_tokens)
        collate._lang_id = lang_id
        dataloader = DataLoader(dataset, batch_size=batch_size, collate_fn=collate.collate_fn,
                                shuffle=False, num_workers=num_workers)

        toks = []
        preds = []
        import tqdm
        for batch in dataloader:  # tqdm.tqdm(dataloader):
            for key in batch:
                if isinstance(batch[key], torch.Tensor):
                    batch[key] = batch[key].to(self._device)

            x_text = batch['x_text']
            y_offset = batch['y_offset'].cpu().numpy()
            y_len = batch['y_len'].cpu().numpy()
            with torch.no_grad():
                y_pred = self.forward(batch)
            y_pred = torch.argmax(y_pred, dim=-1).detach().cpu().numpy()
            for ii in range(len(y_len)):
                ofs = y_offset[ii]
                for jj in range(y_len[ii]):
                    toks.append(x_text[ii][jj])
                    preds.append(y_pred[ii, jj + ofs])

        p_sents = []
        tok_p = ''
        p_mwes = []
        p_sent = []
        p_mwe = []
        for pred, text in zip(preds, toks):
            text = text.replace('▁', '')
            tok_p += text

            if pred == 2 or pred == 3 or pred == 4:
                if tok_p.strip() != '':
                    p_sent.append(tok_p)
                    if pred == 3:
                        p_mwe.append(True)
                    else:
                        p_mwe.append(False)
                tok_p = ''
            if pred == 4:
                if len(p_sent) != 0:
                    p_sents.append(p_sent)
                    p_mwes.append(p_mwe)
                p_sent = []
                p_mwe = []

        if tok_p.strip() != '':
            p_sent.append(tok_p)
            p_mwe.append(False)
        if len(p_sent) != 0:
            p_sents.append(p_sent)
            p_mwes.append(p_mwe)

        d = Document()

        for sent, mwe in zip(p_sents, p_mwes):
            seq = []
            cnt = 0
            spaceafter = "_"
            for w, m in zip(sent, mwe):
                cnt += 1

                seq.append(Word(cnt, w, '_', '_', '_', '_', 0, '_', '_', spaceafter))
                if m:
                    seq[-1].space_after += ';compund'
            s = Sentence(sequence=seq, lang_id=lang_id)

            d.sentences.append(s)
        return d

    def configure_optimizers(self):
        return torch.optim.AdamW(self.parameters())

    def _compute_early_stop(self, res):
        for lang in res:
            if res[lang]["sent"] > self._res[lang]["sent"]:
                self._early_stop_meta_val += 1
                self._res[lang]["sent"] = res[lang]["sent"]
                res[lang]["sent_best"] = True
            if res[lang]["token"] > self._res[lang]["token"]:
                self._early_stop_meta_val += 1
                self._res[lang]["token"] = res[lang]["token"]
                res[lang]["token_best"] = True
        return res

    def _get_device(self):
        if self._lang_emb.weight.device.type == 'cpu':
            return 'cpu'
        return '{0}:{1}'.format(self._lang_emb.weight.device.type, str(self._lang_emb.weight.device.index))

    def _detect_no_space_lang(document: Document):
        seen_spc = 0
        POLL_RANGE = 50
        for ii in range(POLL_RANGE):
            index = random.randint(0, len(document.sentences) - 1)
            text = document.sentences[index].text.strip()
            if ' ' in text:
                seen_spc += 1
        if seen_spc / POLL_RANGE > 0.5:
            return False
        else:
            return True

    class PrintAndSaveCallback(pl.callbacks.Callback):
        def __init__(self, store_prefix):
            super().__init__()
            self.store_prefix = store_prefix

        def on_validation_end(self, trainer, pl_module):
            metrics = trainer.callback_metrics
            epoch = trainer.current_epoch

            # from pprint import pprint
            # pprint(metrics)
            for lang in pl_module._epoch_results:
                res = pl_module._epoch_results[lang]
                if "sent_best" in res:
                    trainer.save_checkpoint(self.store_prefix + "." + lang + ".sent")
                if "token_best" in res:
                    trainer.save_checkpoint(self.store_prefix + "." + lang + ".tok")

            trainer.save_checkpoint(self.store_prefix + ".last")

            s = "{0:30s}\tSENT\tTOKEN".format("Language")
            print("\n\n\t" + s)
            print("\t" + ("=" * (len(s) + 9)))
            for lang in pl_module._language_codes:
                sent = metrics["val/SENT/{0}".format(lang)]
                token = metrics["val/TOKEN/{0}".format(lang)]
                msg = "\t{0:30s}:\t{1:.4f}\t{2:.4f}".format(lang, sent, token)
                print(msg)
            print("\n")


def _conll_eval(gold, pred):
    f = open('tmp_g.txt', 'w')
    for sent in gold:
        for ii in range(len(sent)):
            head = ii
            f.write('{0}\t{1}\t_\t_\t_\t_\t{2}\t_\t_\t_\n'.format(ii + 1, sent[ii], head))
        f.write('\n')
    f.close()

    f = open('tmp_p.txt', 'w')
    for sent in pred:
        for ii in range(len(sent)):
            head = ii
            f.write('{0}\t{1}\t_\t_\t_\t_\t{2}\t_\t_\t_\n'.format(ii + 1, sent[ii], head))
        f.write('\n')
    f.close()
    from _cube.misc.conll18_ud_eval_wrapper import conll_eval
    result = conll_eval('tmp_g.txt', 'tmp_p.txt')
    if result is None:
        return 0, 0
    else:
        return result['Sentences'].f1, result['Tokens'].f1


"""
if __name__ == '__main__':
    from cube.io_utils.misc import ArgParser

    argparser = ArgParser()
    # run argparser
    args = argparser()
    print(args)  # example

    import json

    langs = json.load(open(args.train_file))
    doc_train = Document()
    doc_dev = Document()
    id2lang = {}
    for ii in range(len(langs)):
        lang = langs[ii]
        print(lang[1], ii)
        doc_train.load(lang[1], lang_id=ii)
        doc_dev.load(lang[2], lang_id=ii)
        id2lang[ii] = lang[0]

    # ensure target dir exists
    target = args.store
    i = args.store.rfind("/")
    if i > 0:
        target = args.store[:i]
        os.makedirs(target, exist_ok=True)

    enc = Encodings()
    enc.compute(doc_train, None)
    enc.save('{0}.encodings'.format(args.store))

    config = TokenizerConfig()
    no_space_lang = _detect_no_space_lang(doc_train)
    print("NO_SPACE_LANG = " + str(no_space_lang))
    config.no_space_lang = no_space_lang
    config.lm_model = args.lm_model
    if args.config_file:
        config.load(args.config_file)
        if args.lm_model is not None:
            config.lm_model = args.lm_model
    config.save('{0}.config'.format(args.store))

    # helper = LMHelper(device=args.lm_device, model=config.lm_model)
    # helper.apply(doc_dev)
    # helper.apply(doc_train)
    trainset = TokenizationDataset(doc_train)
    devset = TokenizationDataset(doc_dev, shuffle=False)
    collate = TokenCollate(enc, lm_device=args.lm_device, lm_model=args.lm_model, no_space_lang=config.no_space_lang)
    train_loader = DataLoader(trainset, batch_size=args.batch_size, collate_fn=collate.collate_fn, shuffle=True,
                              num_workers=args.num_workers)
    val_loader = DataLoader(devset, batch_size=args.batch_size, collate_fn=collate.collate_fn,
                            num_workers=args.num_workers)

    model = Tokenizer(config=config, encodings=enc, id2lang=id2lang)

    # training

    early_stopping_callback = EarlyStopping(
        monitor='val/early_meta',
        patience=args.patience,
        verbose=True,
        mode='max'
    )
    if args.gpus == 0:
        acc = 'ddp_cpu'
    else:
        acc = 'ddp'
    trainer = pl.Trainer(
        gpus=args.gpus,
        accelerator=acc,
        num_nodes=1,
        default_root_dir='data/',
        callbacks=[early_stopping_callback, PrintAndSaveCallback(args, id2lang)],
        # limit_train_batches=5,
        # limit_val_batches=2,
    )

    trainer.fit(model, train_loader, val_loader)
"""



import sys


def get_upos(unimorf):
    upos_list = ['N', 'PROPN', 'ADJ', 'PRO', 'CLF', 'ART', 'DET', 'V', 'ADV', 'AUX', 'V.PTCP', 'V.MSDR', 'V.CVB', 'ADP',
                 'COMP', 'CONJ', 'NUM', 'PART', 'INTJ']
    parts = unimorf.split(';')
    for part in parts:
        if part in upos_list:
            return part
    # print(unimorf)
    return "_"


def get_attrs(unimorf, upos):
    parts = unimorf.split(';')
    if upos not in parts:
        print (upos)
    else:
        parts.remove(upos)
    parts = sorted(parts)
    attrs = ''
    for part in parts:
        attrs += part + ';'
    if attrs == '':
        attrs = '_;'
    attrs = attrs[:-1]
    return attrs


def update_sigmorphon(input_file, output_file):
    inpf = open(input_file, 'r')
    outf = open(output_file, 'w')
    for line in inpf:
        parts = line.split('\t')
        if len(parts) < 3:
            outf.write(line)
        else:
            unimorf_label = parts[5]
            upos = get_upos(unimorf_label)
            attrs = get_attrs(unimorf_label, upos)

            o_line = ''
            for part, index in zip(parts, range(len(parts))):
                if index == 3:
                    o_line += upos + '\t'
                elif index == 5:
                    o_line += attrs + '\t'
                else:
                    o_line += part + '\t'

            o_line = o_line[:-1]
            outf.write(o_line)

    inpf.close()
    outf.close()


def display_help():
    sys.stdout.write('Usage: reshape_unimorf.py <input file> <output_file>\n')


if len(sys.argv) != 3:
    display_help()
else:
    update_sigmorphon(sys.argv[1], sys.argv[2])



#
# Authors: Tiberiu Boros, Stefan Daniel Dumitrescu
#
# Copyright (c) 2018 Adobe Systems Incorporated. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import sys
import dynet as dy
import numpy as np
import os
import random
import string
from cube.misc.misc import fopen

from cube.misc.misc import get_eta, pretty_time, log_progress, line_count
from cube.io_utils.conll import ConllEntry
from cube.generic_networks.crf import CRFDecoder


class DummyTokenizer:
    def __init__(self, config, encodings, num_languages=1, dict=None, runtime=False):
        self.model = dy.Model()
        self.trainer = dy.AdamTrainer(self.model)
        self.encodings = encodings
        self.config = config
        self.dict = dict
        self.char_lookup = self.model.add_lookup_parameters((len(self.encodings.char2int), config.char_emb_size))
        self.case_lookup = self.model.add_lookup_parameters((3, 32))
        self.lang_lookup = self.model.add_lookup_parameters((num_languages, config.lang_emb_size))

        self.LAYER_SIZE = 300
        self.NUM_LAYERS = 15
        self.WINDOW_SIZE = 2
        inp_size = config.lang_emb_size + config.lang_emb_size + 32
        if dict is not None:
            inp_size += 1
        self._proj_w = self.model.add_parameters((self.LAYER_SIZE, inp_size))
        self._proj_b = self.model.add_parameters((self.LAYER_SIZE))

        self._fw_gate_w = []
        self._fw_gate_b = []
        self._fw_act_w = []
        self._fw_act_b = []
        self._fw_skip_w = []
        self._fw_skip_b = []
        self._bw_gate_w = []
        self._bw_gate_b = []
        self._bw_act_w = []
        self._bw_act_b = []
        self._bw_skip_w = []
        self._bw_skip_b = []
        inp_size = self.LAYER_SIZE * (self.WINDOW_SIZE + 1)
        for ii in range(self.NUM_LAYERS):
            self._fw_gate_w.append(self.model.add_parameters((self.LAYER_SIZE, inp_size)))
            self._fw_gate_b.append(self.model.add_parameters((self.LAYER_SIZE)))
            self._fw_act_w.append(self.model.add_parameters((self.LAYER_SIZE, inp_size)))
            self._fw_act_b.append(self.model.add_parameters((self.LAYER_SIZE)))
            self._fw_skip_w.append(self.model.add_parameters((self.LAYER_SIZE, inp_size)))
            self._fw_skip_b.append(self.model.add_parameters((self.LAYER_SIZE)))
            self._bw_gate_w.append(self.model.add_parameters((self.LAYER_SIZE, inp_size)))
            self._bw_gate_b.append(self.model.add_parameters((self.LAYER_SIZE)))
            self._bw_act_w.append(self.model.add_parameters((self.LAYER_SIZE, inp_size)))
            self._bw_act_b.append(self.model.add_parameters((self.LAYER_SIZE)))
            self._bw_skip_w.append(self.model.add_parameters((self.LAYER_SIZE, inp_size)))
            self._bw_skip_b.append(self.model.add_parameters((self.LAYER_SIZE)))
            inp_size = self.LAYER_SIZE * (self.WINDOW_SIZE + 1)

        self.label2int = {'B': 0, 'I': 1, 'E': 2, 'S': 3, 'X': 4, 'BM': 5, 'IM': 6, 'EM': 7, 'SM': 8, 'T': 9, 'U': 10,
                          'UM': 11}
        self.label_list = ['B', 'I', 'E', 'S', 'X', 'BM', 'IM', 'EM', 'SM', 'T', 'U', 'UM']
        # self.crf = CRFDecoder(self.model, self.LAYER_SIZE, 100, len(self.label_list))
        self.aux1_softmax_output_w = self.model.add_parameters((len(self.label_list), self.LAYER_SIZE))
        self.aux1_softmax_output_b = self.model.add_parameters((len(self.label_list)))
        self.aux2_softmax_output_w = self.model.add_parameters((len(self.label_list), self.LAYER_SIZE))
        self.aux2_softmax_output_b = self.model.add_parameters((len(self.label_list)))
        self.softmax_output_w = self.model.add_parameters((len(self.label_list), self.LAYER_SIZE * 2))
        self.softmax_output_b = self.model.add_parameters((len(self.label_list)))

    def _make_input(self, seqs):
        chars = []
        tags = []

        for seq in seqs:
            for entry in seq:
                for char_idx in range(len(entry.word)):
                    chars.append(entry.word[char_idx])
                    if len(entry.word) == 1:
                        tags.append('S')
                    elif char_idx == 0:
                        tags.append('B')
                    elif char_idx == len(entry.word) - 1:
                        tags.append('E')
                    else:
                        tags.append('I')
                    if entry.is_compound_entry:
                        tags[-1] = tags[-1] + 'M'

                if "spaceafter=no" not in entry.space_after.lower():
                    chars.append(' ')
            delta = -1
            while tags[delta].startswith('X'):
                delta -= 1  # it should never crash if dataset is ok

            if tags[delta].startswith('S'):
                tags[delta] = 'T'
            else:
                append_m = ''
                if tags[delta].endswith('M'):
                    append_m = 'M'
                tags[delta] = 'U' + append_m
        return chars, tags

    def learn(self, conll_sequences, lang_id=0):
        dy.renew_cg()
        chars, tags = self._make_input(conll_sequences)

        outputs, aux1, aux2 = self._forward(chars, lang_id=lang_id)
        # crf_output = self.crf.tag(outputs)
        tgt_tags = [self.label2int[tag] for tag in tags]

        # loss = self.crf.learn(outputs, tgt_tags) / len(tgt_tags)
        # from ipdb import set_trace
        # set_trace()

        loss_main = dy.esum([-dy.log(dy.pick(output, tgt)) for output, tgt in zip(outputs, tgt_tags)]) / len(tgt_tags)
        loss_aux1 = dy.esum([-dy.log(dy.pick(output, tgt)) for output, tgt in zip(aux1, tgt_tags)]) / len(tgt_tags)
        loss_aux2 = dy.esum([-dy.log(dy.pick(output, tgt)) for output, tgt in zip(aux2, tgt_tags)]) / len(tgt_tags)

        loss = loss_main + 0.5 * loss_aux1 + 0.5 * loss_aux2
        l_val = loss.value() * len(tgt_tags)
        loss.backward()
        self.trainer.update()
        return l_val

    def _batch_forward(self, chars, lang_id, batch_size=1000):
        tags = []
        num_batches = len(chars) // batch_size
        if len(chars) % batch_size != 0:
            num_batches += 1

        for ii in range(num_batches):
            start = ii * batch_size
            stop = ii * batch_size + batch_size
            if stop > len(chars):
                stop = len(chars)

            cbs = stop - start
            offset = 0
            if start > 0:
                start -= self.WINDOW_SIZE * self.NUM_LAYERS
                offset = self.WINDOW_SIZE * self.NUM_LAYERS
            stop = min(len(chars), stop + self.WINDOW_SIZE * self.NUM_LAYERS)

            copy_chars = chars[start:stop]
            dy.renew_cg()
            # from ipdb import set_trace
            # set_trace()
            outputs, _, _ = self._forward(copy_chars, lang_id=lang_id, runtime=True)
            outputs = outputs[offset:offset + cbs]

            new_tags = [np.argmax(pred.npvalue()) for pred in outputs]
            # new_tags = self.crf.tag(outputs)
            for tag in new_tags:
                tags.append(self.label_list[tag])
        return tags

    def tokenize(self, raw_text, lang_id=0):
        # make sequences of approx 2000-4000 chars
        BATCH_SIZE = 1000  # len(raw_text)
        start = 0
        seqs = []
        seq = []
        word_index = 1
        word = ''

        chars = [c for c in raw_text]
        tags = self._batch_forward(chars, batch_size=BATCH_SIZE, lang_id=lang_id)

        for index, char, tag in zip(range(len(tags)), chars, tags):
            if not tag.startswith('X'):
                word += char
            if not tag.startswith('B') and not tag.startswith('X') and not tag.startswith(
                    'I') and word.strip() != '':
                entry = ConllEntry(word_index, word, '_', '_', '_', '_', word_index - 1, '_', '_', '_')
                seq.append(entry)
                word_index += 1
                word = ''
            if tag.startswith('T') or tag.startswith('U'):
                seqs.append(seq)
                seq = []
                word_index = 1
        start += len(chars)

        if word.strip() != '':
            entry = ConllEntry(word_index, word, '_', '_', '_', '_', word_index - 1, '_', '_', '_')
            seq.append(entry)
        if len(seq) != 0:
            seqs.append(seq)

        return seqs

    def _is_known_word(self, chars, index):
        max_win_size = 10
        for zz in range(min(max_win_size, index + 1)):
            word = ''.join(chars[index - zz:index + 1])
            if word in self.dict:
                return dy.scalarInput(1)
        return dy.scalarInput(0)

    def _forward(self, chars, lang_id=0, runtime=True):
        inp_size = self.config.char_emb_size + 32 + self.config.lang_emb_size
        if self.dict is not None:
            inp_size += 1
        inp = [dy.inputVector(np.zeros(inp_size)) for ii in
               range(self.WINDOW_SIZE)]

        lang_emb = self.lang_lookup[lang_id]

        for index, char in zip(range(len(chars)), chars):
            case_emb = self.case_lookup[0]
            if char.lower() == char.upper():
                case_emb = self.case_lookup[1]
            elif char.lower() != char:
                case_emb = self.case_lookup[2]
            if char.lower() in self.encodings.char2int:
                char_emb = self.char_lookup[self.encodings.char2int[char.lower()]]

            else:
                char_emb = self.char_lookup[self.encodings.char2int['<UNK>']]

            if self.dict is None:
                inp.append(dy.concatenate([case_emb, char_emb, lang_emb]))
            else:
                inp.append(dy.concatenate([case_emb, char_emb, lang_emb, self._is_known_word(chars, index)]))

        for ii in range(self.WINDOW_SIZE + 1):
            inp.append(dy.inputVector(np.zeros((inp_size))))

        # outputs = []

        input = [self._proj_w.expr(update=True) * x + self._proj_b.expr(update=True) for x in inp]
        inp_fw = input

        skip_conn_fw = [[] for ii in range(len(chars))]
        for idx, g_w, g_b, a_w, a_b, skip_w, skip_b in zip(range(len(self._fw_gate_w)), self._fw_gate_w,
                                                           self._fw_gate_b,
                                                           self._fw_act_w, self._fw_act_b, self._fw_skip_w,
                                                           self._fw_skip_b):
            new_inp = [dy.inputVector(np.zeros(self.LAYER_SIZE)) for ii in
                       range(self.WINDOW_SIZE)]

            for ii in range(len(chars)):
                hidden = dy.concatenate(inp_fw[ii:ii + self.WINDOW_SIZE + 1])
                skip_conn_fw[ii].append(skip_w.expr(update=True) * hidden + skip_b.expr(update=True))

                act = dy.tanh(a_w.expr(update=True) * hidden + a_b.expr(update=True))
                gate = dy.logistic(g_w.expr(update=True) * hidden + g_b.expr(update=True))
                output = dy.cmult(act, gate) + dy.cmult(1.0 - gate, inp_fw[ii + self.WINDOW_SIZE])

                if not runtime:
                    output = dy.dropout(output, 0.25)
                new_inp.append(output)

            for ii in range(self.WINDOW_SIZE + 1):
                new_inp.append(dy.inputVector(np.zeros((self.LAYER_SIZE))))
            inp_fw = new_inp

        inp_bw = input

        skip_conn_bw = [[] for ii in range(len(chars))]
        for idx, g_w, g_b, a_w, a_b, skip_w, skip_b in zip(range(len(self._bw_gate_w)), self._bw_gate_w,
                                                           self._bw_gate_b,
                                                           self._bw_act_w, self._bw_act_b, self._bw_skip_w,
                                                           self._bw_skip_b):
            new_inp = [dy.inputVector(np.zeros(self.LAYER_SIZE)) for ii in
                       range(self.WINDOW_SIZE)]

            for ii in range(len(chars)):
                hidden = dy.concatenate(inp_bw[ii + self.WINDOW_SIZE: ii + self.WINDOW_SIZE * 2 + 1])
                skip_conn_bw[ii].append(skip_w.expr(update=True) * hidden + skip_b.expr(update=True))

                act = dy.tanh(a_w.expr(update=True) * hidden + a_b.expr(update=True))
                gate = dy.logistic(g_w.expr(update=True) * hidden + g_b.expr(update=True))
                output = dy.cmult(act, gate) + dy.cmult(1.0 - gate, inp_bw[ii + self.WINDOW_SIZE])

                if not runtime:
                    output = dy.dropout(output, 0.25)
                new_inp.append(output)

            for ii in range(self.WINDOW_SIZE + 1):
                new_inp.append(dy.inputVector(np.zeros((self.LAYER_SIZE))))
            inp_bw = new_inp

        outputs_aux1 = [
            dy.softmax(
                self.aux1_softmax_output_w.expr(update=True) * dy.rectify(
                    dy.esum(res)) + self.aux1_softmax_output_b.expr(
                    update=True)) for hidden, res in zip(inp_fw[self.WINDOW_SIZE:], skip_conn_fw)]
        outputs_aux2 = [
            dy.softmax(
                self.aux2_softmax_output_w.expr(update=True) * dy.rectify(
                    dy.esum(res)) + self.aux2_softmax_output_b.expr(
                    update=True)) for hidden, res in zip(inp_bw[self.WINDOW_SIZE:], skip_conn_bw)]

        outputs = [
            dy.softmax(
                self.softmax_output_w.expr(update=True) * dy.rectify(
                    dy.concatenate([dy.esum(resfw), dy.esum(resbw)])) + self.softmax_output_b.expr(
                    update=True)) for hidden, resfw, resbw in
                    zip(inp_fw[self.WINDOW_SIZE:], skip_conn_fw, skip_conn_bw)]
        # outputs = inp[self.window_size:]
        return outputs, outputs_aux1, outputs_aux2

    def save(self, filename):
        self.model.save(filename)

    def load(self, filename):
        self.model.populate(filename)


class CRFTokenizer:
    def __init__(self, config, encodings, num_languages=1, runtime=False):
        self.model = dy.Model()
        self.trainer = dy.AdamTrainer(self.model)
        self.encodings = encodings
        self.config = config
        self.char_lookup = self.model.add_lookup_parameters((len(self.encodings.char2int), config.char_emb_size))
        self.case_lookup = self.model.add_lookup_parameters((3, 32))
        input_size = 32 + config.char_emb_size

        self.lang_lookup = self.model.add_lookup_parameters((num_languages, config.lang_emb_size))
        input_size += config.lang_emb_size

        label_list = ['B', 'I', 'E', 'S', 'X', 'BM', 'IM', 'EM', 'SM', 'T', 'U', 'UM']
        self.label_list = []
        self.label2int = {}
        for upos in self.encodings.upos2int:
            for label in label_list:
                key = label + '_' + upos
                self.label2int[key] = len(self.label2int)
                self.label_list.append(key)

        # self.label2int = {'B': 0, 'I': 1, 'E': 2, 'S': 3, 'X': 4, 'BM': 5, 'IM': 6, 'EM': 7, 'SM': 8, 'T': 9, 'U': 10,
        #                   'UM': 11}
        # self.label_list = ['B', 'I', 'E', 'S', 'X', 'BM', 'IM', 'EM', 'SM', 'T', 'U', 'UM']

        self.lstm_fw = []
        self.lstm_bw = []

        for layer_size in self.config.lstm_layers:
            self.lstm_fw.append(dy.LSTMBuilder(1, input_size, layer_size, self.model))
            self.lstm_bw.append(dy.LSTMBuilder(1, input_size, layer_size, self.model))
            input_size = layer_size * 2 + self.config.lang_emb_size

        self.crf_decoder = CRFDecoder(self.model, input_size, 100, len(self.label2int))

    def _make_input(self, seqs):
        chars = []
        tags = []

        for seq in seqs:
            for entry in seq:
                upos = ''
                for char_idx in range(len(entry.word)):
                    chars.append(entry.word[char_idx])
                    upos = entry.upos
                    if len(entry.word) == 1:
                        tags.append('S')
                    elif char_idx == 0:
                        tags.append('B')
                    elif char_idx == len(entry.word) - 1:
                        tags.append('E')
                    else:
                        tags.append('I')
                    if entry.is_compound_entry:
                        tags[-1] = tags[-1] + 'M'

                    tags[-1] = tags[-1] + '_' + upos
                if "spaceafter=no" not in entry.space_after.lower():
                    chars.append(' ')
                    tags.append('X_' + upos)
            delta = -1
            while tags[delta].startswith('X'):
                delta -= 1  # it should never crash if dataset is ok

            if tags[delta].startswith('S'):
                tags[delta] = 'T_' + upos
            else:
                append_m = ''
                if tags[delta][1] == 'M':
                    append_m = 'M'
                tags[delta] = 'U' + append_m + '_' + upos
        return chars, tags

    def learn(self, conll_sequences, lang_id=0):
        dy.renew_cg()
        chars, tags = self._make_input(conll_sequences)
        # for char, tag in zip(chars, tags):
        #     print(char + "-" + tag)
        #
        # for seq in conll_sequences:
        #     for entry in seq:
        #         print(entry.word + "\t" + entry.space_after)
        # sys.exit(0)
        embs = self._forward(chars, lang_id=lang_id)
        tgt_tags = [self.label2int[tag] for tag in tags]
        loss = self.crf_decoder.learn(embs, tgt_tags)
        l_val = loss.value()
        loss.backward()
        self.trainer.update()
        return l_val

    def tokenize(self, raw_text, lang_id=0):
        # make sequences of approx 2000-4000 chars
        BATCH_SIZE = 2000
        start = 0
        seqs = []
        seq = []
        word_index = 1
        word = ''

        while start < len(raw_text):
            dy.renew_cg()
            stop = start + BATCH_SIZE
            if len(raw_text) - stop < BATCH_SIZE:
                stop = len(raw_text)

            chars = [c for c in raw_text[start:stop]]
            embs = self._forward(chars, lang_id=lang_id)
            tags = self.crf_decoder.tag(embs)

            tmp_tags = [self.label_list[tag] for tag in tags]
            tags = tmp_tags

            for index, char, tag in zip(range(len(tags)), chars, tags):
                if not tag.startswith('X'):
                    word += char
                if not tag.startswith('B') and not tag.startswith('X') and not tag.startswith(
                        'I') and word.strip() != '':
                    entry = ConllEntry(word_index, word, '_', '_', '_', '_', word_index - 1, '_', '_', '_')
                    seq.append(entry)
                    word_index += 1
                    word = ''
                if tag.startswith('T') or tag.startswith('U'):
                    seqs.append(seq)
                    seq = []
                    word_index = 1
            start += len(chars)

        if word.strip() != '':
            entry = ConllEntry(word_index, word, '_', '_', '_', '_', word_index - 1, '_', '_', '_')
            seq.append(entry)
        if len(seq) != 0:
            seqs.append(seq)

        return seqs

    def _forward(self, chars, lang_id=0):
        inp = []

        lang_emb = self.lang_lookup[lang_id]

        for char in chars:
            case_emb = self.case_lookup[0]
            if char.lower() == char.upper():
                case_emb = self.case_lookup[1]
            elif char.lower() != char:
                case_emb = self.case_lookup[2]
            if char.lower() in self.encodings.char2int:
                char_emb = self.char_lookup[self.encodings.char2int[char.lower()]]

            else:
                char_emb = self.char_lookup[self.encodings.char2int['<UNK>']]

            inp.append(dy.concatenate([case_emb, char_emb, lang_emb]))

        for l_fw, l_bw in zip(self.lstm_fw, self.lstm_bw):
            x_fw = l_fw.initial_state().transduce(inp)
            x_bw = l_fw.initial_state().transduce(inp)
            inp = [dy.concatenate([fw, bw, lang_emb]) for fw, bw in zip(x_fw, list(reversed(x_bw)))]

        return inp

    def save(self, filename):
        self.model.save(filename)

    def load(self, filename):
        self.model.populate(filename)


class TieredTokenizer:
    def __init__(self, config, encodings, embeddings, runtime=False):
        self.config = config
        self.word_embeddings = embeddings
        self.encodings = encodings

        self.modelSS = dy.Model()
        self.modelTok = dy.Model()
        self.trainerSS = dy.AdamTrainer(self.modelSS, alpha=2e-3, beta_1=0.9, beta_2=0.9)
        self.trainerTok = dy.AdamTrainer(self.modelTok, alpha=2e-3, beta_1=0.9, beta_2=0.9)

        # sentence split model
        from cube.generic_networks.wrappers import CNN, CNNConvLayer, CNNPoolingLayer
        from cube.generic_networks.utils import orthonormal_VanillaLSTMBuilder
        # character-level-embeddings
        self.SS_char_lookup = self.modelSS.add_lookup_parameters(
            (len(self.encodings.char2int), self.config.ss_char_embeddings_size))
        self.SS_char_lookup_casing = self.modelSS.add_lookup_parameters((3, 5))  # lower, upper N/A
        self.SS_char_lookup_special = self.modelSS.add_lookup_parameters((2, self.config.ss_char_embeddings_size + 5))
        # lstm-peek network
        if runtime:
            self.SS_peek_lstm = dy.VanillaLSTMBuilder(self.config.ss_peek_lstm_layers,
                                                      self.config.ss_char_embeddings_size + 5,
                                                      self.config.ss_peek_lstm_size, self.modelSS)
        else:
            self.SS_peek_lstm = orthonormal_VanillaLSTMBuilder(self.config.ss_peek_lstm_layers,
                                                               self.config.ss_char_embeddings_size + 5,
                                                               self.config.ss_peek_lstm_size, self.modelSS)
        layer_is = self.config.ss_peek_lstm_size
        self.SS_aux_softmax_peek_w = self.modelSS.add_parameters((2, layer_is))
        self.SS_aux_softmax_peek_b = self.modelSS.add_parameters((2))
        if runtime:
            self.SS_lstm = dy.VanillaLSTMBuilder(self.config.ss_lstm_layers, self.config.ss_char_embeddings_size + 5,
                                                 self.config.ss_lstm_size,
                                                 self.modelSS)
        else:
            self.SS_lstm = orthonormal_VanillaLSTMBuilder(self.config.ss_lstm_layers,
                                                          self.config.ss_char_embeddings_size + 5,
                                                          self.config.ss_lstm_size,
                                                          self.modelSS)

        self.SS_aux_softmax_prev_w = self.modelSS.add_parameters((2, self.config.ss_lstm_size))
        self.SS_aux_softmax_prev_b = self.modelSS.add_parameters((2))

        # post MLP and softmax
        self.SS_mlp_w = []
        self.SS_mlp_b = []
        layer_is = self.config.ss_lstm_size + self.config.ss_peek_lstm_size
        for layer in self.config.ss_mlp_layers:
            self.SS_mlp_w.append(self.modelSS.add_parameters((layer, layer_is)))
            self.SS_mlp_b.append(self.modelSS.add_parameters((layer)))
            layer_is = layer

        self.SS_mlp_softmax_w = self.modelSS.add_parameters((2, layer_is))
        self.SS_mlp_softmax_b = self.modelSS.add_parameters((2))

        # tokenization model
        self.TOK_char_lookup = self.modelTok.add_lookup_parameters(
            (len(self.encodings.char2int), self.config.tok_char_embeddings_size))
        self.TOK_char_lookup_casing = self.modelTok.add_lookup_parameters((3, 5))  # lower, upper N/A
        self.TOK_char_lookup_special = self.modelTok.add_lookup_parameters(
            (2, self.config.tok_char_embeddings_size + 5))
        self.TOK_word_lookup = self.modelTok.add_lookup_parameters(
            (len(self.encodings.word2int), self.config.tok_word_embeddings_size))

        self.TOK_word_embeddings_special = self.modelTok.add_lookup_parameters(
            (2, self.word_embeddings.word_embeddings_size))

        self.TOK_word_proj_w = self.modelTok.add_parameters(
            (self.config.tok_word_embeddings_size, self.word_embeddings.word_embeddings_size))
        self.TOK_word_peek_proj_w = self.modelTok.add_parameters(
            (self.config.tok_word_embeddings_size, self.word_embeddings.word_embeddings_size))
        # lstm networks
        if runtime:
            self.TOK_backward_lstm = dy.VanillaLSTMBuilder(self.config.tok_char_peek_lstm_layers,
                                                           self.config.tok_char_embeddings_size + 5,
                                                           self.config.tok_char_peek_lstm_size, self.modelTok)
            self.TOK_forward_lstm = dy.VanillaLSTMBuilder(self.config.tok_char_lstm_layers,
                                                          self.config.tok_char_embeddings_size + 5,
                                                          self.config.tok_char_lstm_size, self.modelTok)
            self.TOK_word_lstm = dy.VanillaLSTMBuilder(self.config.tok_word_lstm_layers,
                                                       self.config.tok_word_embeddings_size,
                                                       self.config.tok_word_lstm_size,
                                                       self.modelTok)
        else:
            self.TOK_backward_lstm = orthonormal_VanillaLSTMBuilder(self.config.tok_char_peek_lstm_layers,
                                                                    self.config.tok_char_embeddings_size + 5,
                                                                    self.config.tok_char_peek_lstm_size, self.modelTok)
            self.TOK_forward_lstm = orthonormal_VanillaLSTMBuilder(self.config.tok_char_lstm_layers,
                                                                   self.config.tok_char_embeddings_size + 5,
                                                                   self.config.tok_char_lstm_size, self.modelTok)
            self.TOK_word_lstm = orthonormal_VanillaLSTMBuilder(self.config.tok_word_lstm_layers,
                                                                self.config.tok_word_embeddings_size,
                                                                self.config.tok_word_lstm_size,
                                                                self.modelTok)

        self.TOK_mlp_w = []
        self.TOK_mlp_b = []
        layer_input = self.config.tok_word_lstm_size + self.config.tok_char_lstm_size + self.config.tok_char_peek_lstm_size + 2 + self.config.tok_word_embeddings_size
        for layer_size in self.config.tok_mlp_layers:
            self.TOK_mlp_w.append(self.modelTok.add_parameters((layer_size, layer_input)))
            self.TOK_mlp_b.append(self.modelTok.add_parameters((layer_size)))
            layer_input = layer_size

        self.TOK_softmax_w = self.modelTok.add_parameters((2, layer_input))
        self.TOK_softmax_b = self.modelTok.add_parameters((2))
        self.TOK_softmax_peek_w = self.modelTok.add_parameters((2, self.config.tok_char_peek_lstm_size))
        self.TOK_softmax_peek_b = self.modelTok.add_parameters((2))
        self.TOK_softmax_prev_w = self.modelTok.add_parameters((2, self.config.tok_char_lstm_size))
        self.TOK_softmax_prev_b = self.modelTok.add_parameters((2))

        self.losses = []
        self.losses_tok = []

    def start_batch(self):
        self.losses = []
        self.losses_tok = []
        dy.renew_cg()

    def end_batch(self):
        total_loss_val = 0
        if len(self.losses) > 0:
            total_loss_ss = dy.esum(self.losses)
            total_loss_val += total_loss_ss.value()
            total_loss_ss.backward()
            self.trainerSS.update()

        if len(self.losses_tok):
            total_loss_tok = dy.esum(self.losses_tok)
            total_loss_val += total_loss_tok.value()
            total_loss_tok.backward()
            self.trainerTok.update()
        return total_loss_val

    def _predict_tok(self, seq, y_gold=None, runtime=False):
        x_list = []
        offset = 1

        word_is_known = dy.inputVector([1.0, 0.0])
        word_is_unknown = dy.inputVector([0.0, 1.0])

        for char in seq:
            lookup_char = char.lower()
            import re
            lookup_char = re.sub('\d', '0', lookup_char)
            if lookup_char in self.encodings.char2int:
                char_emb = self.TOK_char_lookup[self.encodings.char2int[lookup_char]]
            else:
                char_emb = self.TOK_char_lookup[self.encodings.char2int['<UNK>']]

            if char.lower() == char and char.upper() == char:
                casing_emb = self.TOK_char_lookup_casing[0]  # does not support casing
            elif char.lower() == char:
                casing_emb = self.TOK_char_lookup_casing[1]  # is lowercased
            else:
                casing_emb = self.TOK_char_lookup_casing[2]  # is uppercased

            x = dy.concatenate([char_emb, casing_emb])
            x_list.append(x)

        for _ in range(offset):
            x_list.append(self.TOK_char_lookup_special[1])

        aux_softmax_output_peek = []
        aux_softmax_output_prev = []
        softmax_output = []

        if runtime:
            self.TOK_forward_lstm.set_dropouts(0, 0)
            self.TOK_backward_lstm.set_dropouts(0, 0)
            self.TOK_word_lstm.set_dropouts(0, 0)
        else:
            self.TOK_forward_lstm.set_dropouts(self.config.tok_char_lstm_dropout, self.config.tok_char_lstm_dropout)
            self.TOK_backward_lstm.set_dropouts(self.config.tok_char_peek_lstm_dropout,
                                                self.config.tok_char_peek_lstm_dropout)
            self.TOK_word_lstm.set_dropouts(self.config.tok_word_lstm_dropout, self.config.tok_word_lstm_dropout)

        fw_out = self.TOK_forward_lstm.initial_state().transduce(x_list)
        bw_out = list(reversed(self.TOK_backward_lstm.initial_state().transduce(reversed(x_list))))
        word_lstm = self.TOK_word_lstm.initial_state().add_input(
            dy.inputVector([0] * self.config.tok_word_embeddings_size))
        word = ""
        for index in range(len(seq)):
            word += seq[index]
            aux_softmax_output_prev.append(
                dy.softmax(self.TOK_softmax_prev_w.expr(update=True) * fw_out[index] + self.TOK_softmax_prev_b.expr(
                    update=True)))
            aux_softmax_output_peek.append(
                dy.softmax(self.TOK_softmax_peek_w.expr(update=True) * bw_out[index] + self.TOK_softmax_peek_b.expr(
                    update=True)))

            word_state = word_is_unknown
            peek_emb, found = self.word_embeddings.get_word_embeddings(word.strip())
            if found:
                word_state = word_is_known
                peek_emb = self.TOK_word_peek_proj_w.expr(update=True) * dy.inputVector(peek_emb)
            else:
                peek_emb = self.TOK_word_peek_proj_w.expr(update=True) * self.TOK_word_embeddings_special[0]

            if word.strip().lower() in self.encodings.word2int:
                word_state = word_is_known
                peek_hol = self.TOK_word_lookup[self.encodings.word2int[word.strip().lower()]]
            else:
                peek_hol = self.TOK_word_lookup[self.encodings.word2int['<UNK>']]

            hidden = dy.concatenate(
                [fw_out[index], bw_out[index], word_lstm.output(), word_state, dy.tanh(peek_hol + peek_emb)])
            for w, b, dropout in zip(self.TOK_mlp_w, self.TOK_mlp_b, self.config.tok_mlp_dropouts):
                hidden = dy.tanh(w.expr(update=True) * hidden + b.expr(update=True))
                if not runtime:
                    hidden = dy.dropout(hidden, dropout)

            softmax_output.append(
                dy.softmax(self.TOK_softmax_w.expr(update=True) * hidden + self.TOK_softmax_b.expr(update=True)))
            must_split = False
            if not runtime:
                if y_gold[index] == "S" or y_gold[index] == "SX":
                    must_split = True
            elif np.argmax(softmax_output[-1].npvalue()) == 1:
                must_split = True

            if must_split:
                emb, found = self.word_embeddings.get_word_embeddings(word.strip())
                if not found:
                    emb = self.TOK_word_embeddings_special[1]
                else:
                    emb = dy.inputVector(emb)

                word = word.lower().strip()
                if word in self.encodings.word2int:
                    hol = self.TOK_word_lookup[self.encodings.word2int[word]]
                else:
                    hol = self.TOK_word_lookup[self.encodings.word2int['<UNK>']]

                emb = self.TOK_word_proj_w.expr(update=True) * emb

                word_lstm = word_lstm.add_input(dy.tanh(emb + hol))
                word = ""

        return softmax_output, aux_softmax_output_prev, aux_softmax_output_peek

    def learn_tok(self, ix, iy, aux_softmax_weight=0.2):
        losses = []
        # we must split into sentences before learning tokenization - otherwise, the data will be inconsistent with the runtime version
        start = 0
        import copy
        x = copy.deepcopy(ix)
        y = copy.deepcopy(iy)
        while True:
            if "SX" in y:
                index_of_ss = y.index("SX")
            else:
                break

            y_prediction, y_prediction_aux1, y_prediction_aux2 = self._predict_tok(x[start:index_of_ss + 1],
                                                                                   y_gold=y[start:index_of_ss + 1],
                                                                                   runtime=False)
            for y_real, y_pred, y_aux1, y_aux2 in zip(y, y_prediction, y_prediction_aux1, y_prediction_aux2):
                if y_real == "SX" or y_real == "S":
                    losses.append(-dy.log(dy.pick(y_pred, 1)))
                    losses.append(-dy.log(dy.pick(y_aux1, 1)) * aux_softmax_weight)
                    losses.append(-dy.log(dy.pick(y_aux2, 1)) * aux_softmax_weight)
                else:
                    losses.append(-dy.log(dy.pick(y_pred, 0)))
                    losses.append(-dy.log(dy.pick(y_aux1, 0)) * aux_softmax_weight)
                    losses.append(-dy.log(dy.pick(y_aux2, 0)) * aux_softmax_weight)
            x = x[index_of_ss + 1:]
            y = y[index_of_ss + 1:]

        loss = dy.esum(losses)
        self.losses_tok.append(loss)

    def learn_ss(self, x, y, aux_softmax_weight=0.2):
        losses = []
        y_prediction, y_prediction_aux1, y_prediction_aux2 = self._predict_ss(x, runtime=False)
        for y_real, y_pred, y_aux1, y_aux2 in zip(y, y_prediction, y_prediction_aux1, y_prediction_aux2):
            if y_real == "SX":
                losses.append(-dy.log(dy.pick(y_pred, 1)))
                losses.append(-dy.log(dy.pick(y_aux1, 1)) * aux_softmax_weight)
                losses.append(-dy.log(dy.pick(y_aux2, 1)) * aux_softmax_weight)
            else:
                losses.append(-dy.log(dy.pick(y_pred, 0)))
                losses.append(-dy.log(dy.pick(y_aux1, 0)) * aux_softmax_weight)
                losses.append(-dy.log(dy.pick(y_aux2, 0)) * aux_softmax_weight)

        loss = dy.esum(losses)
        self.losses.append(loss)

    def _get_tokens(self, input_string, space_after_end_of_sentence=True):
        # print("\n")
        # print(input_string)
        tokens = []
        y_pred, _, _ = self._predict_tok(input_string, runtime=True)
        index = 0
        w = ""

        for i in range(len(input_string)):
            w += input_string[i]
            if np.argmax(y_pred[i].npvalue()) == 1:
                if w.strip() != "":
                    index += 1
                    space_after = "SpaceAfter=No"
                    if i < len(input_string) - 1:
                        if input_string[i + 1] in string.whitespace:
                            space_after = "_"
                    entry = ConllEntry(index, str(w).strip(), '_', "_", "_", "_", 0, "_", "_",
                                       space_after=space_after)
                    tokens.append(entry)
                    w = ""

        if w.strip() != "":
            index += 1
            entry = ConllEntry(index, str(w).strip(), '_', "_", "_", "_", 0, "_", "_", "")
            tokens.append(entry)

        # set SpaceAfter=No property of last token
        if len(tokens) > 0:
            tokens[-1].space_after = "SpaceAfter=No" if space_after_end_of_sentence == False else "_"

        return tokens

    def tokenize(self, input_string):
        batch_size = 1000
        import sys
        if sys.version_info[0] == 2:
            input_string = unicode(input_string, 'utf-8')
        else:
            import copy
            input_string = copy.deepcopy(input_string)

        sequences = []
        num_chars = 0
        last_proc = 0
        sz = len(input_string)

        w = ""
        while len(input_string) > 0:
            current_string = input_string[:min(len(input_string), batch_size + 1)]
            """
            proc = num_chars * 100 / sz
            while last_proc + 5 < proc:
                last_proc += 5
                sys.stdout.write(" " + str(last_proc))
                sys.stdout.flush()
            """

            dy.renew_cg()
            y_pred, _, _ = self._predict_ss(current_string)

            last_ss_break = -1
            last_checked_index = -1
            if len(current_string) == batch_size:
                current_string = current_string[:-100]

            for y, char, index in zip(y_pred, current_string, range(len(current_string))):
                w += char
                if np.argmax(y.npvalue()) == 1:
                    space_after_end_of_sentence = False
                    if index < len(
                            input_string) - 1:  # compare with input_string not with current_string for whitespace after current sentence
                        if input_string[index + 1] in string.whitespace:
                            space_after_end_of_sentence = True
                    seq = self._get_tokens(w.strip(), space_after_end_of_sentence=space_after_end_of_sentence)
                    if seq:
                        sequences.append(seq)
                    w = ""
                    last_ss_break = index
                last_checked_index = index

            if last_ss_break == -1:  # no sentence break applied
                last_ss_break = last_checked_index
            else:
                w = ""

            input_string = input_string[last_ss_break + 1:]

            num_chars += last_ss_break

        if w.strip() != "":
            space_after_end_of_sentence = False
            if w[-1] in string.whitespace:
                space_after_end_of_sentence = True
            seq = self._get_tokens(w.strip(), space_after_end_of_sentence=space_after_end_of_sentence)
            sequences.append(seq)

        return sequences

    def save_ss(self, filename):
        self.modelSS.save(filename)

    def save_tok(self, filename):
        self.modelTok.save(filename)

    def load(self, filename):
        self.modelTok.populate(filename + "-tok.bestAcc")
        self.modelSS.populate(filename + "-ss.bestAcc")

    def _predict_ss(self, seq, runtime=True):
        x_list = []
        offset = self.config.ss_char_peek_count

        for char in seq:
            lookup_char = char.lower()
            import re
            lookup_char = re.sub('\d', '0', lookup_char)
            if lookup_char in self.encodings.char2int:
                char_emb = self.SS_char_lookup[self.encodings.char2int[lookup_char]]
            else:
                char_emb = self.SS_char_lookup[self.encodings.char2int['<UNK>']]

            if char.lower() == char and char.upper() == char:
                casing_emb = self.SS_char_lookup_casing[0]  # does not support casing
            elif char.lower() == char:
                casing_emb = self.SS_char_lookup_casing[1]  # is lowercased
            else:
                casing_emb = self.SS_char_lookup_casing[2]  # is uppercased

            x = dy.concatenate([char_emb, casing_emb])
            x_list.append(x)

        for _ in range(offset):
            x_list.append(self.SS_char_lookup_special[1])

        aux_softmax_output_peek = []
        aux_softmax_output_prev = []
        softmax_output = []

        # forward pass should not be a problem
        if runtime:
            self.SS_lstm.set_dropouts(0, 0)
            self.SS_peek_lstm.set_dropouts(0, 0)
        else:
            self.SS_lstm.set_dropouts(self.config.ss_lstm_dropout, self.config.ss_lstm_dropout)
            self.SS_peek_lstm.set_dropouts(self.config.ss_peek_lstm_dropout, self.config.ss_peek_lstm_dropout)
        lstm_fw = self.SS_lstm.initial_state()

        for cIndex in range(len(seq)):
            peek_chars = x_list[cIndex:cIndex + self.config.ss_char_peek_count + 1]
            peek_out = self.SS_peek_lstm.initial_state().transduce(reversed(peek_chars))[-1]

            aux_softmax_output_peek.append(
                dy.softmax(self.SS_aux_softmax_peek_w.expr(update=True) * peek_out + self.SS_aux_softmax_peek_b.expr(
                    update=True)))

            lstm_fw = lstm_fw.add_input(x_list[cIndex])
            lstm_out = lstm_fw.output()
            aux_softmax_output_prev.append(
                dy.softmax(self.SS_aux_softmax_prev_w.expr(update=True) * lstm_out + self.SS_aux_softmax_prev_b.expr(
                    update=True)))
            hidden = dy.concatenate([lstm_out, peek_out])
            for w, b, dropout in zip(self.SS_mlp_w, self.SS_mlp_b, self.config.ss_mlp_dropouts):
                hidden = dy.tanh(w.expr(update=True) * hidden + b.expr(update=True))
                if not runtime:
                    hidden = dy.dropout(hidden, dropout)
            softmax_output.append(
                dy.softmax(self.SS_mlp_softmax_w.expr(update=True) * hidden + self.SS_mlp_softmax_b.expr(update=True)))

        return softmax_output, aux_softmax_output_peek, aux_softmax_output_prev


class BDRNNTokenizer:

    def __init__(self, config, encodings, embeddings, runtime=False):
        # INTERNAL PARAMS ###################################################        
        self.config = config
        self.encodings = encodings
        self.word_embeddings = embeddings
        self.config.char_vocabulary_size = len(encodings.characters)
        self.decoder_output_class_count = 3  # O S SX
        self.decoder_output_i2c = {}
        self.decoder_output_i2c[0] = "O"
        self.decoder_output_i2c[1] = "S"
        self.decoder_output_i2c[2] = "SX"
        self.decoder_output_c2i = {}
        self.decoder_output_c2i["O"] = 0
        self.decoder_output_c2i["S"] = 1
        self.decoder_output_c2i["SX"] = 2

        # NETWORK ###########################################################    
        self.model = dy.Model()
        self.trainer = dy.AdamTrainer(self.model)
        self.trainer.set_sparse_updates(False)

        # EMBEDDING SPECIAL TOKENS
        self.word_embeddings_special = self.model.add_lookup_parameters(
            (2, self.word_embeddings.word_embeddings_size))  # [0] = UNK, [1] = SENTENCE START

        # ENCODER-CHAR        
        self.char_embeddings = self.model.add_lookup_parameters(
            (self.config.char_vocabulary_size, self.config.char_embedding_size))
        # self.next_chars_embedding = self.model.add_lookup_parameters(
        #    (self.config.char_vocabulary_size, self.config.next_chars_embedding_size))

        self.char_embeddings_punctuation = self.model.add_lookup_parameters(
            (self.config.char_generic_feature_vocabulary_size, self.config.char_generic_feature_embedding_size))
        self.char_embeddings_whitespace = self.model.add_lookup_parameters(
            (self.config.char_generic_feature_vocabulary_size, self.config.char_generic_feature_embedding_size))
        self.char_embeddings_uppercase = self.model.add_lookup_parameters(
            (self.config.char_generic_feature_vocabulary_size, self.config.char_generic_feature_embedding_size))
        self.encoder_char_input_size = self.config.char_embedding_size + 3 * self.config.char_generic_feature_embedding_size
        if runtime:
            self.encoder_char_lstm1_fw_builder = dy.VanillaLSTMBuilder(1, self.encoder_char_input_size,
                                                                       self.config.encoder_char_lstm_size, self.model)

            self.encoder_char_lstm2_bw_builder = dy.VanillaLSTMBuilder(1,
                                                                       self.config.next_chars_embedding_size + 3 * self.config.char_generic_feature_embedding_size,
                                                                       self.config.encoder_char_lstm_size, self.model)
            self.encoder_word_lstm_builder = dy.VanillaLSTMBuilder(1, self.word_embeddings.word_embeddings_size,
                                                                   self.config.encoder_word_lstm_size, self.model)
        else:
            from generic_networks.utils import orthonormal_VanillaLSTMBuilder
            self.encoder_char_lstm1_fw_builder = orthonormal_VanillaLSTMBuilder(1, self.encoder_char_input_size,
                                                                                self.config.encoder_char_lstm_size,
                                                                                self.model)

            self.encoder_char_lstm2_bw_builder = orthonormal_VanillaLSTMBuilder(1,
                                                                                self.config.next_chars_embedding_size + 3 * self.config.char_generic_feature_embedding_size,
                                                                                self.config.encoder_char_lstm_size,
                                                                                self.model)
            self.encoder_word_lstm_builder = orthonormal_VanillaLSTMBuilder(1,
                                                                            self.word_embeddings.word_embeddings_size,
                                                                            self.config.encoder_word_lstm_size,
                                                                            self.model)

        # ENCODER-WORD

        # self.att_w1 = self.model.add_parameters((
        #     self.config.next_chars_embedding_size + self.config.char_generic_feature_embedding_size * 3,
        #     self.config.encoder_char_lstm_size))
        # self.att_w2 = self.model.add_parameters((
        #     self.config.next_chars_embedding_size + self.config.char_generic_feature_embedding_size * 3,
        #     self.config.encoder_char_lstm_size))
        # self.att_v = self.model.add_parameters(
        #     (1, self.config.next_chars_embedding_size + self.config.char_generic_feature_embedding_size * 3))

        # DECODER

        self.holisticWE = self.model.add_lookup_parameters(
            (len(encodings.word2int), self.word_embeddings.word_embeddings_size))

        self.decoder_input_size = 2 * self.config.encoder_char_lstm_size + self.config.encoder_word_lstm_size + self.word_embeddings.word_embeddings_size

        self.decoder_hiddenW = self.model.add_parameters((self.config.decoder_hidden_size, self.decoder_input_size))
        self.decoder_hiddenB = self.model.add_parameters((self.config.decoder_hidden_size))
        self.decoder_outputW = self.model.add_parameters(
            (self.decoder_output_class_count, self.config.decoder_hidden_size))
        self.decoder_outputB = self.model.add_parameters((self.decoder_output_class_count))

        self.aux_softmax_char_peek_w = self.model.add_parameters(
            (self.decoder_output_class_count, self.config.encoder_char_lstm_size))
        self.aux_softmax_char_peek_b = self.model.add_parameters((self.decoder_output_class_count))

        self.aux_softmax_char_hist_w = self.model.add_parameters(
            (self.decoder_output_class_count, self.config.encoder_char_lstm_size))
        self.aux_softmax_char_hist_b = self.model.add_parameters((self.decoder_output_class_count))
        print("done")

    def learn(self, X, y):  # X is a list of symbols, y is a list of labels for each symbol
        softmax, softmax_aux1, softmax_aux2 = self._predict(X, y)

        losses = []

        for i in range(len(softmax)):
            loss = -dy.log(dy.pick(softmax[i], self.decoder_output_c2i[y[i]]))
            losses.append(loss)
            loss_aux = -dy.log(dy.pick(softmax_aux1[i], self.decoder_output_c2i[y[i]])) * 0.2
            losses.append(loss_aux)
            loss_aux = -dy.log(dy.pick(softmax_aux2[i], self.decoder_output_c2i[y[i]])) * 0.2
            losses.append(loss_aux)
        l = dy.esum(losses)
        loss = l.value()
        l.backward()
        self.trainer.update()
        return loss

    def tokenize(self,
                 input_string):  # input string is a single string that can contain several sentences, output will be conllu-format list of sentences
        import sys
        if sys.version_info[0] == 2:
            uni_string = unicode(input_string, 'utf-8')  # because reasons
        else:
            uni_string = input_string
        offset = 0
        sentences = []
        last_proc = 0
        while offset < len(uni_string):
            proc = (offset + 1) * 100 / len(uni_string)

            while last_proc + 5 < proc:
                last_proc += 5
                sys.stdout.write(" " + str(last_proc))
                sys.stdout.flush()

            # print("Total len = "+str(len(uni_string)))
            # print("Current offset = "+str(offset))
            window = 0
            while True:  # extend window until we find an end of sentence (SX)
                window += self.config.tokenize_maximum_sequence_length
                X = uni_string[offset:min(len(uni_string), offset + window)]
                # print("    X len = "+str(len(X)))
                softmax, _, _ = self._predict(X)
                # print("    Softmax len = "+str(len(softmax)))
                # convert to labels
                labels = [self.decoder_output_i2c[np.argmax(s.npvalue())] for s in softmax]
                # print("    Predicted label len = "+str(len(labels)))
                if "SX" in labels:
                    break
                elif offset + len(labels) >= len(
                        uni_string):  # maybe we reached end of input_string without an SX, then exit as well
                    break
            offset += len(labels)

            # create sentence from labels
            sentence = []
            word = ""
            cnt = 1
            # with fopen("log.txt","a") as log:
            # log.write("\n\n")
            for i in range(len(labels)):
                # log.write("["+X[i].encode('utf-8')+"] "+labels[i]+" w=["+word.encode('utf-8')+"]\n")
                if "O" in labels[i]:
                    word = word + X[i]
                if "S" in labels[i]:
                    if X[i] in string.whitespace:  # if whitespace, skip
                        if word != "":
                            entry = ConllEntry(index=cnt, word=word.decode('utf-8'), lemma="_", upos="_", xpos="_",
                                               attrs="_", head="0",
                                               label="_", deps="_", space_after="_")
                            # log.write("   New ERROR incomplete entry ["+word.encode('utf-8')+"]\n")
                            sentence.append(entry)
                            cnt += 1
                            word = ""
                        continue
                    word += X[i]
                    space_after = "SpaceAfter=No"
                    if i < len(X) - 1:
                        if X[i + 1] in string.whitespace:
                            space_after = "_"
                    entry = ConllEntry(index=cnt, word=word.decode('utf-8'), lemma="_", upos="_", xpos="_", attrs="_",
                                       head="0",
                                       label="_", deps="_", space_after=space_after)
                    # log.write("   New entry ["+word.encode('utf-8')+"]\n")
                    sentence.append(entry)
                    cnt += 1
                    word = ""
            sentences.append(sentence)

            # for entry in sentence:
            #    print(" \t Word :"+entry.word)

        return sentences

    def _attend(self, input_vectors, state):
        w1 = self.att_w1.expr(update=True)
        w2 = self.att_w2.expr(update=True)
        v = self.att_v.expr(update=True)
        attention_weights = []

        w2dt = w2 * state.h()[-1]
        for input_vector in input_vectors:
            attention_weight = v * dy.tanh(w1 * input_vector + w2dt)
            attention_weights.append(attention_weight)

        attention_weights = dy.softmax(dy.concatenate(attention_weights))

        output_vectors = dy.esum(
            [vector * attention_weight for vector, attention_weight in zip(input_vectors, attention_weights)])

        return output_vectors

    def _predict(self, X, y=None):  # X is a list of symbols, y is a list of labels for each symbol
        dy.renew_cg()
        runtime = True if y is None else False
        debug = False
        if debug: print("Predict sequence start, len = " + str(len(X)))

        # ENCODER-CHAR
        lstm1_forward = self.encoder_char_lstm1_fw_builder.initial_state()

        # ENCODER-WORD         
        encoder_word_lstm = self.encoder_word_lstm_builder.initial_state()
        encoder_word_lstm = encoder_word_lstm.add_input(self.word_embeddings_special[1])  # [1] = SENTENCE START
        encoder_word_output = encoder_word_lstm.output()

        # PARTIAL-EMBEDDING
        partial_word = self.word_embeddings_special[1]  # [1] = SENTENCE START

        # DECODER
        prediction_output = []
        softmax_dy = []
        softmax_aux_peek = []
        softmax_aux_hist = []
        index_last_word_start = 0

        zeroVecArray_next_chars_embedding = dy.inputVector(
            [0] * (self.config.next_chars_embedding_size + self.config.char_generic_feature_embedding_size * 3))

        for char_index in range(len(X)):  # for each character in X
            if debug: print(" Char index " + str(char_index) + " of " + str(len(X)) + " = \"" + X[char_index] + "\"")

            # ENCODER-CHAR
            encoder_char_input = self._encoder_char_encode_input(X[char_index:char_index + 1],
                                                                 runtime)  # encode only the current symbol
            # print(encoder_char_input)
            lstm1_forward = lstm1_forward.add_input(encoder_char_input[0])
            encoder_char_output = lstm1_forward.output()

            # PARTIAL-EMBEDDING
            if index_last_word_start == 0:
                partial_word = self.word_embeddings_special[1]  # [1] = sentence_start
            else:
                word = ""
                for i in range(index_last_word_start, char_index + 1):
                    word += X[i]
                    # word = word.decode('utf-8')
                partial_word = self._encoder_word_encode_input(word)
                if debug: print("   partial word : [" + word + "]")

            # # NEXT-CHARS MLP
            next_chars = self._next_chars_encode_input(
                X[char_index:min(len(X), char_index + self.config.next_chars_window_size)], runtime)
            remaining = self.config.next_chars_window_size - (len(X) - char_index)
            # # if debug: print("      *NEXT-CHARS: remaining: "+str(remaining)+" current len : "+str(len(next_chars)))
            for r in range(remaining):
                next_chars.append(zeroVecArray_next_chars_embedding)

            next_chars = self.encoder_char_lstm2_bw_builder.initial_state().transduce(reversed(next_chars))
            next_chars = next_chars[-1]  # self._attend(next_chars, lstm1_forward)

            softmax_aux_peek.append(
                dy.softmax(
                    self.aux_softmax_char_peek_w.expr(update=True) * next_chars + self.aux_softmax_char_peek_b.expr(
                        update=True)))
            softmax_aux_hist.append(dy.softmax(
                self.aux_softmax_char_hist_w.expr(
                    update=True) * encoder_char_output + self.aux_softmax_char_hist_b.expr(update=True)))

            # dropout at feature-set level:
            # if runtime:
            decoder_input = dy.concatenate([encoder_char_output, encoder_word_output, partial_word,
                                            next_chars])  # embeddings concatenation of char and word
            # decoder_input = dy.concatenate([encoder_char_output, next_chars])
            if not runtime:
                decoder_input = dy.dropout(decoder_input, self.config.dropout_rate)

            decoder_hidden = dy.tanh(
                self.decoder_hiddenW.expr(update=True) * decoder_input + self.decoder_hiddenB.expr(update=True))
            if not runtime:
                decoder_hidden = dy.dropout(decoder_hidden, self.config.dropout_rate)

            softmax_output = dy.softmax(
                self.decoder_outputW.expr(update=True) * decoder_hidden + self.decoder_outputB.expr(update=True))
            softmax_dy.append(softmax_output)
            if runtime:
                prediction = self.decoder_output_i2c[np.argmax(softmax_output.npvalue())]
            else:  # learn
                prediction = y[
                    char_index]  # so force prediction to be = gold-standard, so we have correct word splits

            if debug: print("    Predicted output: " + prediction)

            if "S" in prediction:
                # compose word
                word = ""
                for i in range(index_last_word_start, char_index + 1):
                    word += X[i]

                index_last_word_start = char_index + 1
                # get ENCODER-WORD output
                current_word_embedding = self._encoder_word_encode_input(word)
                encoder_word_lstm = encoder_word_lstm.add_input(current_word_embedding)
                encoder_word_output = encoder_word_lstm.output()
                # RESET the CHAR LSTM
                lstm1_forward = self.encoder_char_lstm1_fw_builder.initial_state()

            if "X" in prediction:  # stop everything
                break

        if runtime and debug:
            raw_input("STEP")
        return softmax_dy, softmax_aux_peek, softmax_aux_hist

    def _encoder_word_encode_input(self, word):
        embedding, _ = self.word_embeddings.get_word_embeddings(word)
        word = word.lower()
        if word in self.encodings.word2int:
            hol_we = self.holisticWE[self.encodings.word2int[word]]
        else:
            hol_we = self.holisticWE[self.encodings.word2int['<UNK>']]
        if embedding is not None:
            return dy.inputVector(embedding) + hol_we
        else:  # UNKNOWN WORD
            return self.word_embeddings_special[0] + hol_we  # UNK

    def _next_chars_encode_input(self, X, runtime=True):
        encoded = []
        for i in range(len(X)):
            char = X[i]
            char_lower = X[i].lower()
            import re
            char_lower = re.sub('\d', '0', char_lower)
            char = re.sub('\d', '0', char)

            if char_lower in self.encodings.char2int:
                embedding = self.char_embeddings[self.encodings.char2int[char_lower]]
            else:
                embedding = self.char_embeddings[
                    self.encodings.char2int["<UNK>"]]  # dy.inputVector([0]*self.config.next_chars_embedding_size)

            all_info = [embedding]
            if char_lower != char:
                all_info.append(self.char_embeddings_uppercase[1])
            else:
                all_info.append(self.char_embeddings_uppercase[0])

            if char_lower in string.whitespace:
                all_info.append(self.char_embeddings_whitespace[1])
            else:
                all_info.append(self.char_embeddings_whitespace[0])

            # is punctuation
            if char_lower in string.punctuation:
                all_info.append(self.char_embeddings_punctuation[1])
            else:
                all_info.append(self.char_embeddings_punctuation[0])

            encoded.append(dy.concatenate(all_info))
        return encoded

    def _encoder_char_encode_input(self, X, runtime=True):
        encoded = []
        for char_index in range(len(X)):
            embeddings = []
            # char itself
            char = X[char_index]
            char_lower = char.lower()
            import re
            char_lower = re.sub('\d', '0', char_lower)
            char = re.sub('\d', '0', char)
            if char_lower not in self.encodings.char2int:
                # print (" UNK CHAR: ["+char+"]")
                char_lower = "<UNK>"
            embeddings.append(self.char_embeddings[self.encodings.char2int[char_lower]])
            # is uppercase
            if char.lower() != char:
                embeddings.append(self.char_embeddings_uppercase[1])
            else:
                embeddings.append(self.char_embeddings_uppercase[0])
            # is whitespace
            if char in string.whitespace:
                embeddings.append(self.char_embeddings_whitespace[1])
            else:
                embeddings.append(self.char_embeddings_whitespace[0])
            # is punctuation
            if char in string.punctuation:
                embeddings.append(self.char_embeddings_punctuation[1])
            else:
                embeddings.append(self.char_embeddings_punctuation[0])

            input = dy.concatenate(embeddings)
            # if not runtime:
            #    input = dy.noise(input,self.config.encoder_char_input_noise)
            encoded.append(input)
        return encoded

    def save(self, path):
        self.model.save(path)

    def load(self, path):
        self.model.populate(path)

    def _argmax(self, probs):
        max_index = 0
        for zz in range(len(probs)):
            if probs[zz] > probs[max_index]:
                max_index = zz
        return max_index



#
# Author: Tiberiu Boros
#
# Copyright (c) 2019 Adobe Systems Incorporated. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import torch
import torch.nn as nn
import torch.nn.functional as F
import random
from torch.nn.utils.rnn import PackedSequence
from typing import *
import pytorch_lightning as pl


class LinearNorm(pl.LightningModule):
    def __init__(self, in_dim, out_dim, bias=True, w_init_gain='linear'):
        super(LinearNorm, self).__init__()
        self.linear_layer = nn.Linear(in_dim, out_dim, bias=bias)

        torch.nn.init.xavier_normal_(
            self.linear_layer.weight,
            gain=torch.nn.init.calculate_gain(w_init_gain))

    def forward(self, x):
        return self.linear_layer(x)


class MLP(nn.Module):
    def __init__(self, in_dim, out_dim, hid_dim=500, dropout=0.33, hid_func=torch.tanh):
        super().__init__()
        self._h1 = LinearNorm(in_dim, hid_dim)
        self._h2 = LinearNorm(hid_dim, out_dim)
        self._dropout = dropout
        self._hid_func = hid_func

    def forward(self, x):
        h = torch.dropout(self._hid_func(self._h1(x)), self._dropout, self.training)
        o = self._h2(h)
        return o


class Encoder(pl.LightningModule):
    def __init__(self, input_type, input_size, input_emb_dim, enc_hid_dim, dropout, nn_type=nn.GRU,
                 num_layers=2, ext_conditioning=0):
        super().__init__()
        assert (input_type == 'int' or input_type == 'float')
        if input_type == 'float':
            assert (input_size == input_emb_dim)
        self.input_type = input_type
        self.input_dim = input_size
        self.emb_dim = input_emb_dim
        self.enc_hid_dim = enc_hid_dim

        if self.input_type == 'int':
            self.embedding = nn.Sequential(nn.Embedding(input_size, input_emb_dim), nn.Dropout(dropout))
        else:
            self.embedding = nn.Dropout(dropout)
        if nn_type == VariationalLSTM:
            self.rnn = nn_type(input_emb_dim + ext_conditioning, enc_hid_dim, bidirectional=True, num_layers=1,
                               dropoutw=dropout, dropouto=dropout, dropouti=dropout, batch_first=True)
            self.dropout = nn.Identity()
        else:
            self.rnn = nn_type(input_emb_dim + ext_conditioning, enc_hid_dim, bidirectional=True, num_layers=1,
                               batch_first=True)
            self.dropout = nn.Dropout(dropout)

        if num_layers > 1:
            top_layers = []
            for ii in range(num_layers - 1):
                if nn_type == VariationalLSTM:
                    top_layers.append(
                        nn_type(enc_hid_dim * 2 + ext_conditioning, enc_hid_dim, bidirectional=True, num_layers=1,
                                batch_first=True, dropoutw=dropout, dropouto=dropout, dropouti=dropout))
                else:
                    top_layers.append(
                        nn_type(enc_hid_dim * 2 + ext_conditioning, enc_hid_dim, bidirectional=True, num_layers=1,
                                batch_first=True))

            self.top_layers = nn.ModuleList(top_layers)
        else:
            self.top_layers = None

    def forward(self, src, conditioning=None):
        embedded = self.embedding(src)

        if conditioning is not None:
            # conditioning = conditioning.permute(0, 1)
            conditioning = conditioning.unsqueeze(1)
            conditioning = conditioning.repeat(1, src.shape[1], 1)
            embedded = torch.cat((embedded, conditioning), dim=2)

        outputs, hidden = self.rnn(embedded)
        if self.top_layers is not None:
            for rnn_layer in self.top_layers:
                if conditioning is not None:
                    outputs, hidden = rnn_layer(torch.cat((self.dropout(outputs), conditioning), dim=2))
                else:
                    outputs, hidden = rnn_layer(self.dropout(outputs))
        if isinstance(hidden, list) or isinstance(hidden, tuple):  # we have a LSTM
            hidden = hidden[1]
        hidden = torch.cat((hidden[-1, :, :], hidden[0, :, :]), dim=1)

        return outputs, hidden


class BilinearAttention(nn.Module):
    def __init__(self, dim1, dim2):
        super().__init__()
        self.biliniar = nn.Bilinear(dim1, dim2, 1)
        self.linear = nn.Linear(dim1 + dim2, 1)

    def forward(self, query, keys):
        query = query.unsqueeze(1).repeat(1, keys.shape[1], 1)
        biliniar = self.biliniar(query, keys)
        h = torch.cat([query, keys], dim=-1)
        liniar = self.linear(h)
        return (liniar + biliniar).squeeze(2)


class Attention(pl.LightningModule):
    def __init__(self, enc_hid_dim, dec_hid_dim, att_proj_size=100):
        super().__init__()

        self.enc_hid_dim = enc_hid_dim
        self.dec_hid_dim = dec_hid_dim

        # self.attn = LinearNorm((enc_hid_dim * 2) + dec_hid_dim, dec_hid_dim)
        self.attn = ConvNorm(enc_hid_dim * 2 + dec_hid_dim, att_proj_size, kernel_size=5,
                             w_init_gain='tanh')
        self.v = nn.Parameter(torch.rand(att_proj_size))

    def forward(self, hidden, encoder_outputs, return_logsoftmax=False):
        # hidden = [batch size, dec hid dim]
        # encoder_outputs = [src sent len, batch size, enc hid dim * 2]
        batch_size = encoder_outputs.shape[0]
        src_len = encoder_outputs.shape[1]
        # repeat encoder hidden state src_len times
        hidden = hidden.unsqueeze(1).repeat(1, src_len, 1)
        encoder_outputs = encoder_outputs
        # hidden = [batch size, src sent len, dec hid dim]
        # encoder_outputs = [batch size, src sent len, enc hid dim * 2]
        energy = torch.dropout(
            torch.tanh(self.attn(torch.cat((hidden, encoder_outputs), dim=2).transpose(1, 2)).transpose(1, 2)), 0.1,
            self.training)
        energy = energy.transpose(1, 2)
        # energy = [batch size, src sent len, dec hid dim]
        # energy = [batch size, dec hid dim, src sent len]
        # v = [dec hid dim]
        v = self.v.repeat(batch_size, 1).unsqueeze(1)
        # v = [batch size, 1, dec hid dim]
        attention = torch.bmm(v, energy).squeeze(1)
        # attention= [batch size, src len]
        if return_logsoftmax:
            return F.log_softmax(attention, dim=1)
        else:
            return F.softmax(attention, dim=1)


class Decoder(pl.LightningModule):
    def __init__(self, output_dim, emb_dim, enc_hid_dim, dec_hid_dim, dropout, attention, nn_type=nn.GRU, num_layers=2):
        super().__init__()

        self.emb_dim = emb_dim
        self.enc_hid_dim = enc_hid_dim
        self.dec_hid_dim = dec_hid_dim
        self.output_dim = output_dim
        self.dropout = dropout
        self.attention = attention

        self.embedding = nn.Embedding(output_dim, emb_dim)
        self.rnn = nn_type((enc_hid_dim * 2) + emb_dim, dec_hid_dim, num_layers=num_layers)
        self.out = LinearNorm((enc_hid_dim * 2) + dec_hid_dim + emb_dim, output_dim)
        self.dropout = nn.Dropout(dropout)

    def forward(self, input, hidden, encoder_outputs):
        input = input.unsqueeze(0)
        embedded = self.dropout(self.embedding(input))
        a = self.attention(hidden, encoder_outputs)
        a = a.unsqueeze(1)
        encoder_outputs = encoder_outputs
        weighted = torch.bmm(a, encoder_outputs)
        weighted = weighted
        rnn_input = torch.cat((embedded, weighted), dim=2)
        output, hidden = self.rnn(rnn_input, hidden.unsqueeze(0))

        assert (output == hidden).all()
        embedded = embedded.squeeze(0)
        output = output.squeeze(0)
        weighted = weighted.squeeze(0)
        output = self.out(torch.cat((output, weighted, embedded), dim=1))
        return output, hidden.squeeze(0)


class Seq2Seq(pl.LightningModule):
    def __init__(self, encoder, decoder, device):
        super().__init__()

        self.encoder = encoder
        self.decoder = decoder
        self.device = device

    def forward(self, src, trg, teacher_forcing_ratio=0.5):
        # src = [src sent len, batch size]
        # trg = [trg sent len, batch size]
        # teacher_forcing_ratio is probability to use teacher forcing
        # e.g. if teacher_forcing_ratio is 0.75 we use teacher forcing 75% of the time
        batch_size = src.shape[1]
        max_len = trg.shape[0]
        trg_vocab_size = self.decoder.output_dim
        # tensor to store decoder outputs
        outputs = torch.zeros(max_len, batch_size, trg_vocab_size).to(self.device)
        # encoder_outputs is all hidden states of the input sequence, back and forwards
        # hidden is the final forward and backward hidden states, passed through a linear layer
        encoder_outputs, hidden = self.encoder(src)
        # first input to the decoder is the <sos> tokens
        output = trg[0, :]
        for t in range(1, max_len):
            output, hidden = self.decoder(output, hidden, encoder_outputs)
            outputs[t] = output
            teacher_force = random.random() < teacher_forcing_ratio
            top1 = output.max(1)[1]
            output = (trg[t] if teacher_force else top1)

        return outputs


# The code is adapted from https://github.com/keitakurita/Better_LSTM_PyTorch - no pip package is provided so we are
# cloning the code for testing

class VariationalDropout(pl.LightningModule):
    """
    Applies the same dropout mask across the temporal dimension
    See https://arxiv.org/abs/1512.05287 for more details.
    Note that this is not applied to the recurrent activations in the LSTM like the above paper.
    Instead, it is applied to the inputs and outputs of the recurrent layer.
    """

    def __init__(self, dropout: float, batch_first: Optional[bool] = False):
        super().__init__()
        self.dropout = dropout
        self.batch_first = batch_first

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        if not self.training or self.dropout <= 0.:
            return x

        is_packed = isinstance(x, PackedSequence)
        if is_packed:
            x, batch_sizes = x
            max_batch_size = int(batch_sizes[0])
        else:
            batch_sizes = None
            if self.batch_first:
                max_batch_size = x.size(0)
            else:
                max_batch_size = x.size(1)

        # Drop same mask across entire sequence
        if self.batch_first:
            m = x.new_empty(max_batch_size, 1, x.size(2), requires_grad=False).bernoulli_(1 - self.dropout)
        else:
            m = x.new_empty(1, max_batch_size, x.size(2), requires_grad=False).bernoulli_(1 - self.dropout)
        x = x.masked_fill(m == 0, 0) / (1 - self.dropout)

        if is_packed:
            return PackedSequence(x, batch_sizes)
        else:
            return x


class VariationalLSTM(pl.LightningModule):
    def __init__(self, *args, dropouti: float = 0.,
                 dropoutw: float = 0., dropouto: float = 0.,
                 batch_first=True, unit_forget_bias=True, **kwargs):
        super().__init__(*args, **kwargs, batch_first=batch_first)
        self.unit_forget_bias = unit_forget_bias
        self.dropoutw = dropoutw
        self.input_drop = VariationalDropout(dropouti,
                                             batch_first=batch_first)
        self.output_drop = VariationalDropout(dropouto,
                                              batch_first=batch_first)
        self._init_weights()

    def _init_weights(self):
        """
        Use orthogonal init for recurrent layers, xavier uniform for input layers
        Bias is 0 except for forget gate
        """
        for name, param in self.named_parameters():
            if "weight_hh" in name:
                nn.init.orthogonal_(param.data)
            elif "weight_ih" in name:
                nn.init.xavier_uniform_(param.data)
            elif "bias" in name and self.unit_forget_bias:
                nn.init.zeros_(param.data)
                param.data[self.hidden_size:2 * self.hidden_size] = 1

    def _drop_weights(self):
        for name, param in self.named_parameters():
            if "weight_hh" in name:
                getattr(self, name).data = \
                    torch.nn.functional.dropout(param.data, p=self.dropoutw,
                                                training=self.training).contiguous()

    def forward(self, input, hx=None):
        self._drop_weights()
        self.flatten_parameters()
        input = self.input_drop(input)
        seq, state = super().forward(input, hx=hx)
        return self.output_drop(seq), state


class ConvNorm(pl.LightningModule):
    def __init__(self, in_channels, out_channels, kernel_size=1, stride=1,
                 padding=None, dilation=1, bias=True, w_init_gain='linear'):
        super(ConvNorm, self).__init__()
        if padding is None:
            assert (kernel_size % 2 == 1)
            padding = int(dilation * (kernel_size - 1) / 2)

        self.conv = torch.nn.Conv1d(in_channels, out_channels,
                                    kernel_size=kernel_size, stride=stride,
                                    padding=padding, dilation=dilation,
                                    bias=bias)

        torch.nn.init.xavier_normal_(
            self.conv.weight, gain=torch.nn.init.calculate_gain(w_init_gain))

    def forward(self, signal):
        conv_signal = self.conv(signal)
        return conv_signal


# class WordGram(pl.LightningModule):
#     def __init__(self, num_chars: int, num_langs: int, num_filters=512, char_emb_size=256, case_emb_size=32,
#                  lang_emb_size=32, num_layers=3):
#         super(WordGram, self).__init__()
#         NUM_FILTERS = num_filters
#         self._num_filters = NUM_FILTERS
#         self._lang_emb = nn.Embedding(num_langs + 1, lang_emb_size)
#         self._tok_emb = nn.Embedding(num_chars + 1, char_emb_size)
#         self._case_emb = nn.Embedding(4, case_emb_size)
#         self._num_layers = num_layers
#
#         convolutions_char = []
#         cs_inp = char_emb_size + lang_emb_size + case_emb_size
#         for _ in range(num_layers):
#             conv_layer = nn.Sequential(
#                 ConvNorm(cs_inp,
#                          NUM_FILTERS,
#                          kernel_size=5, stride=1,
#                          padding=2,
#                          dilation=1, w_init_gain='tanh'),
#                 nn.BatchNorm1d(NUM_FILTERS))
#             convolutions_char.append(conv_layer)
#             cs_inp = NUM_FILTERS // 2 + lang_emb_size
#         self._convolutions_char = nn.ModuleList(convolutions_char)
#
#         self._rnn = nn.LSTM(NUM_FILTERS // 2, NUM_FILTERS // 2, num_layers=2)
#         self._pre_out = LinearNorm(NUM_FILTERS // 2, NUM_FILTERS // 2)
#
#     def forward(self, x_char, x_case, x_lang, x_mask, x_word_len):
#         x_char = self._tok_emb(x_char)
#         x_case = self._case_emb(x_case)
#         x_lang = self._lang_emb(x_lang)
#
#         x = torch.cat([x_char, x_case], dim=-1)
#         x = x.permute(0, 2, 1)
#         x_lang = x_lang.unsqueeze(1).repeat(1, x_case.shape[1], 1).permute(0, 2, 1)
#         half = self._num_filters // 2
#         count = 0
#         res = None
#         skip = None
#         for conv in self._convolutions_char:
#             count += 1
#             drop = self.training
#             if count >= len(self._convolutions_char):
#                 drop = False
#             if skip is not None:
#                 x = x + skip
#
#             x = torch.cat([x, x_lang], dim=1)
#             conv_out = conv(x)
#             tmp = torch.tanh(conv_out[:, :half, :]) * torch.sigmoid((conv_out[:, half:, :]))
#             if res is None:
#                 res = tmp
#             else:
#                 res = res + tmp
#             skip = tmp
#             x = torch.dropout(tmp, 0.1, drop)
#         x = x + res
#         x = x.permute(0, 2, 1)
#         x = torch.flip(x, dims=[1])
#         out, _ = self._rnn(x)
#         norm = out[:, -1, :]
#
#         return torch.tanh(self._pre_out(norm))
#
#     def _get_device(self):
#         if self._lang_emb.weight.device.type == 'cpu':
#             return 'cpu'
#         return '{0}:{1}'.format(self._lang_emb.weight.device.type, str(self._lang_emb.weight.device.index))
#
#     def save(self, path):
#         torch.save(self.state_dict(), path)
#
#     def load(self, path):
#         self.load_state_dict(torch.load(path, map_location='cpu'))


class WordGram(pl.LightningModule):
    def __init__(self, num_chars: int, num_langs: int, num_filters=512, char_emb_size=256, case_emb_size=32,
                 lang_emb_size=32, num_layers=3):
        super(WordGram, self).__init__()
        NUM_FILTERS = num_filters
        self._num_filters = NUM_FILTERS
        self._lang_emb = nn.Embedding(num_langs + 1, lang_emb_size)
        self._tok_emb = nn.Embedding(num_chars + 1, char_emb_size)
        self._case_emb = nn.Embedding(4, case_emb_size)
        self._num_layers = num_layers
        convolutions_char = []
        cs_inp = char_emb_size + lang_emb_size + case_emb_size
        for _ in range(num_layers):
            conv_layer = nn.Sequential(
                ConvNorm(cs_inp,
                         NUM_FILTERS,
                         kernel_size=5, stride=1,
                         padding=2,
                         dilation=1, w_init_gain='tanh'),
                nn.BatchNorm1d(NUM_FILTERS))
            convolutions_char.append(conv_layer)
            cs_inp = NUM_FILTERS // 2 + lang_emb_size
        self._convolutions_char = nn.ModuleList(convolutions_char)
        self._pre_out = LinearNorm(NUM_FILTERS // 2, NUM_FILTERS // 2)

    def forward(self, x_char, x_case, x_lang, x_mask, x_word_len):
        x_char = self._tok_emb(x_char)
        x_case = self._case_emb(x_case)
        x_lang = self._lang_emb(x_lang)

        x = torch.cat([x_char, x_case], dim=-1)
        x = x.permute(0, 2, 1)
        x_lang = x_lang.unsqueeze(1).repeat(1, x_case.shape[1], 1).permute(0, 2, 1)
        half = self._num_filters // 2
        count = 0
        res = None
        skip = None
        for conv in self._convolutions_char:
            count += 1
            drop = self.training
            if count >= len(self._convolutions_char):
                drop = False
            if skip is not None:
                x = x + skip

            x = torch.cat([x, x_lang], dim=1)
            conv_out = conv(x)
            tmp = torch.tanh(conv_out[:, :half, :]) * torch.sigmoid((conv_out[:, half:, :]))
            if res is None:
                res = tmp
            else:
                res = res + tmp
            skip = tmp
            x = torch.dropout(tmp, 0.1, drop)
        x = x + res
        x = x.permute(0, 2, 1)
        x = x * x_mask.unsqueeze(2)
        pre = torch.sum(x, dim=1, dtype=torch.float)
        norm = pre / torch.clip(x_word_len.unsqueeze(1), min=1)
        return torch.tanh(self._pre_out(norm))

    def _get_device(self):
        if self._lang_emb.weight.device.type == 'cpu':
            return 'cpu'
        return '{0}:{1}'.format(self._lang_emb.weight.device.type, str(self._lang_emb.weight.device.index))

    def save(self, path):
        torch.save(self.state_dict(), path)

    def load(self, path):
        self.load_state_dict(torch.load(path, map_location='cpu'))


'''
Code adapted from https://github.com/stanfordnlp/stanza/
'''


class PairwiseBilinear(nn.Module):
    ''' A bilinear module that deals with broadcasting for efficient memory usage.
    Input: tensors of sizes (N x L1 x D1) and (N x L2 x D2)
    Output: tensor of size (N x L1 x L2 x O)'''

    def __init__(self, input1_size, input2_size, output_size, bias=True):
        super().__init__()

        self.input1_size = input1_size
        self.input2_size = input2_size
        self.output_size = output_size

        self.weight = nn.Parameter(torch.Tensor(input1_size, input2_size, output_size))
        self.bias = nn.Parameter(torch.Tensor(output_size)) if bias else 0

    def forward(self, input1, input2):
        input1_size = list(input1.size())
        input2_size = list(input2.size())
        output_size = [input1_size[0], input1_size[1], input2_size[1], self.output_size]

        # ((N x L1) x D1) * (D1 x (D2 x O)) -> (N x L1) x (D2 x O)
        intermediate = torch.mm(input1.view(-1, input1_size[-1]),
                                self.weight.view(-1, self.input2_size * self.output_size))
        # (N x L2 x D2) -> (N x D2 x L2)
        input2 = input2.transpose(1, 2)
        # (N x (L1 x O) x D2) * (N x D2 x L2) -> (N x (L1 x O) x L2)
        output = intermediate.view(input1_size[0], input1_size[1] * self.output_size, input2_size[2]).bmm(input2)
        # (N x (L1 x O) x L2) -> (N x L1 x L2 x O)
        output = output.view(input1_size[0], input1_size[1], self.output_size, input2_size[1]).transpose(2, 3)

        return output


class BiaffineScorer(nn.Module):
    def __init__(self, input1_size, input2_size, output_size):
        super().__init__()
        self.W_bilin = nn.Bilinear(input1_size + 1, input2_size + 1, output_size)

        self.W_bilin.weight.data.zero_()
        self.W_bilin.bias.data.zero_()

    def forward(self, input1, input2):
        input1 = torch.cat([input1, input1.new_ones(*input1.size()[:-1], 1)], len(input1.size()) - 1)
        input2 = torch.cat([input2, input2.new_ones(*input2.size()[:-1], 1)], len(input2.size()) - 1)
        return self.W_bilin(input1, input2)


class PairwiseBiaffineScorer(nn.Module):
    def __init__(self, input1_size, input2_size, output_size):
        super().__init__()
        self.W_bilin = PairwiseBilinear(input1_size + 1, input2_size + 1, output_size)

        self.W_bilin.weight.data.zero_()
        self.W_bilin.bias.data.zero_()

    def forward(self, input1, input2):
        input1 = torch.cat([input1, input1.new_ones(*input1.size()[:-1], 1)], len(input1.size()) - 1)
        input2 = torch.cat([input2, input2.new_ones(*input2.size()[:-1], 1)], len(input2.size()) - 1)
        return self.W_bilin(input1, input2)


class DeepBiaffine(nn.Module):
    def __init__(self, input1_size, input2_size, hidden_size, output_size, hidden_func=F.relu, dropout=0,
                 pairwise=True):
        super().__init__()
        self.W1 = nn.Linear(input1_size, hidden_size)
        self.W2 = nn.Linear(input2_size, hidden_size)
        self.hidden_func = hidden_func
        if pairwise:
            self.scorer = PairwiseBiaffineScorer(hidden_size, hidden_size, output_size)
        else:
            self.scorer = BiaffineScorer(hidden_size, hidden_size, output_size)
        self.dropout = nn.Dropout(dropout)

    def forward(self, input1, input2):
        return self.scorer(self.dropout(self.hidden_func(self.W1(input1))),
                           self.dropout(self.hidden_func(self.W2(input2))))



#
# Author: Tiberiu Boros
#
# Copyright (c) 2018 Adobe Systems Incorporated. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import sys
import os
import time

def fopen (filename, mode="r"):
    if sys.version_info[0] == 2:                    
        return open(filename, mode)
    else:                    
        if "b" in mode.lower():
            return open(filename, mode)
        else:
            return open(filename, mode, encoding="utf-8")

# return ETA in seconds
def get_eta(progress, total, time_delta, granularity=2):
    percent_done = float(progress) / float(total)
    return "-" if percent_done <= 0 else pretty_time(seconds=float(time_delta) * (1. / percent_done - percent_done),
                                                     granularity=granularity)


# pretty print time (like: 3d:12h:10m:2s)
def pretty_time(seconds, granularity=2):
    intervals = (('w', 604800), ('d', 86400), ('h', 3600), ('m', 60), ('s', 1))
    result = []
    seconds = int(seconds)
    for name, count in intervals:
        value = seconds // count
        if value:
            seconds -= value * count
            result.append("{}{}".format(value, name))
    return ':'.join(result[:granularity])


def log_progress(output_base, task_name, best_epoch, best_training_acc=0., best_dev_acc=0., other=None):
    filename = output_base + ".log"
    with fopen(filename, "w") as f:
        if other != None:
            if isinstance(other, list):
                for i in range(len(other)):
                    f.write(other[i] + "\n")
        f.write("\n")
        f.write("task_name=" + task_name + "\n")
        f.write("best_epoch=" + str(best_epoch) + "\n")
        f.write("best_training_acc=" + str(best_training_acc) + "\n")
        f.write("best_dev_acc=" + str(best_dev_acc) + "\n")
        """if other != None:
            for key in other:
                f.write(key+"="+str(other["key"])+"\n")"""


def line_count(filename):
    f = fopen(filename)
    lines = 0
    buf_size = 1024 * 1024
    read_f = f.read  # loop optimization
    buf = read_f(buf_size)
    while buf:
        lines += buf.count('\n')
        buf = read_f(buf_size)
    return lines



#
# Author: Tiberiu Boros
#
# Copyright (c) 2018 Adobe Systems Incorporated. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import sys
import io
from cube.misc.misc import fopen


class Dataset:
    def __init__(self, file=None, lang_id=0):
        if file is not None:
            sys.stdout.write("Reading " + file + "... ")
            sys.stdout.flush()
            with fopen(file, "r") as f:
                lines = f.readlines()

            self.sequences = self._make_sequences(lines, lang_id=lang_id)
            sys.stdout.write("found " + str(len(self.sequences)) + " sequences\n")

        self.sequences = []

    def load_language(self, file, lang_id, ignore_compound=False):
        sys.stdout.write("Reading " + file + "... ")
        sys.stdout.flush()
        with fopen(file, "r") as f:
            lines = f.readlines()

        ns = self._make_sequences(lines, lang_id=lang_id, ignore_compound=ignore_compound)
        for [seq, l_id] in ns:
            self.sequences.append([seq, l_id])
        sys.stdout.write("found " + str(len(ns)) + " sequences\n")

    def _make_sequences(self, lines, lang_id=0, ignore_compound=False):
        sequences = []
        in_sequence = False
        seq = []
        for line in lines:
            line = line.replace("\n", "")
            line = line.replace("\r", "")
            if (not line.startswith("#") or in_sequence) and line != '':
                parts = line.split("\t")
                s = ConllEntry(parts[0], parts[1], parts[2], parts[3], parts[4], parts[5], parts[6], parts[7], parts[8],
                               parts[9])
                if not ignore_compound or not s.is_compound_entry:
                    seq.append(s)
                in_sequence = True
            elif line == "":
                in_sequence = False
                if len(seq) > 0:
                    sequences.append([seq, lang_id])
                seq = []

        return sequences

    def write(self, filename):
        with fopen(filename, 'w') as file:
            for sequence in self.sequences:
                for entry in sequence:
                    file.write(str(entry))
                    file.write("\n")
                file.write("\n")

    def write_stdout(self):
        import sys
        file = sys.stdout
        for sequence in self.sequences:
            for entry in sequence:
                file.write(str(entry))
                file.write("\n")
            file.write("\n")

    def write_as_string(self):
        return "".join("".join(str(entry)+"\n" for entry in sequence)+"\n" for sequence in self.sequences)


class Encodings:
    def __init__(self):
        return ""


class ConllEntry:
    def __init__(self, index, word, lemma, upos, xpos, attrs, head, label, deps, space_after):
        self.index, self.is_compound_entry = self._int_try_parse(index)
        self.word = word
        self.lemma = lemma
        self.upos = upos
        self.xpos = xpos
        self.attrs = attrs
        self.head, _ = self._int_try_parse(head)
        self.label = label
        self.deps = deps
        self.space_after = space_after

    def _int_try_parse(self, value):
        try:
            return int(value), False
        except ValueError:
            return value, True

    def __repr__(self):
        return "\t".join([str(self.index), self.word if isinstance(self.word, str) else self.word.encode('utf-8'), self.lemma if isinstance(self.lemma, str) else self.lemma.encode('utf-8'), self.upos, self.xpos, self.attrs, str(self.head), self.label, self.deps, self.space_after])



#
# Author: Tiberiu Boros
#
# Copyright (c) 2018 Adobe Systems Incorporated. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from collections import namedtuple
import numpy as np
import sys

Arc = namedtuple('Arc', ('tail', 'weight', 'head'))


class GreedyDecoder:
    def _valid(self, arc, tree):
        # just one head
        for sa in tree:
            if sa.tail == arc.tail:
                return False
        stack = [arc.head]
        pos = 0
        used = [False] * len(tree)
        while pos < len(stack):
            for zz in range(len(tree)):
                if tree[zz].tail == stack[pos] and not used[zz]:
                    used[zz] = True
                    stack.append(tree[zz].head)
                    if tree[zz].head == arc.tail:
                        return False
            pos += 1
            # print pos,len(stack)
        return True

    def _get_sort_key(self, item):
        return item.weight

    def _greedy_tree(self, arcs):
        arcs = sorted(arcs, key=self._get_sort_key, reverse=True)
        # print arcs
        final_tree = []
        for index in range(len(arcs)):
            if self._valid(arcs[index], final_tree):
                final_tree.append(arcs[index])
                # print arcs[index]
        return final_tree

    def _make_ordered_list(self, tree, nWords):
        lst = [0] * nWords  # np.zeros(nWords)
        for arc in tree:
            # arc = tree[index]
            tail = arc.tail
            head = arc.head
            lst[tail] = head
        return lst[1:]

    def decode(self, score, lens):
        best_tree_list = []
        for ii in range(score.shape[0]):
            # norm_score = score[ii, :lens[ii], :lens[ii]]
            norm_score = np.zeros((lens[ii] + 1, lens[ii] + 1))
            for wii in range(lens[ii]):
                for wjj in range(lens[ii] + 1):
                    norm_score[wii + 1, wjj] = score[ii, wii, wjj]
            nWords = norm_score.shape[0]  # len(norm_score)
            g = []
            for iSrc in range(1, nWords):
                for iDst in range(1, nWords):
                    if iDst != iSrc:
                        a = Arc(iSrc, norm_score[iSrc][iDst], iDst)
                        g.append(a)
            tree = self._greedy_tree(g)
            best_tree = self._make_ordered_list(tree, nWords)
            best_tree_list.append(best_tree)
        return best_tree_list



import sys
sys.path.append('')
import os, yaml
os.environ["TOKENIZERS_PARALLELISM"] = "false"
import pytorch_lightning as pl
import torch.nn as nn
import torch.nn.functional as F
import torch
from torch.utils.data import DataLoader
from cube.io_utils.objects import Document
from cube.io_utils.encodings import Encodings
from cube.io_utils.config import TaggerConfig
from cube.networks.modules import ConvNorm, LinearNorm, MLP
from cube.networks.utils import MorphoCollate, MorphoDataset, unpack, mask_concat
from cube.networks.modules import WordGram

class Tagger(pl.LightningModule):
    def __init__(self, config: TaggerConfig, encodings: Encodings, language_codes: [] = None, ext_word_emb=0):
        super().__init__()
        self._device = "cpu"  # default
        self._config = config
        self._encodings = encodings
        if not isinstance(ext_word_emb, list):
            ext_word_emb = [ext_word_emb]
        self._ext_word_emb = ext_word_emb

        self._word_net = WordGram(len(encodings.char2int), num_langs=encodings.num_langs + 1,
                                  num_filters=config.char_filter_size, char_emb_size=config.char_emb_size,
                                  lang_emb_size=config.lang_emb_size, num_layers=config.char_layers)
        self._zero_emb = nn.Embedding(1, config.char_filter_size // 2)
        self._num_langs = encodings.num_langs
        self._language_codes = language_codes

        conv_layers = []
        cs_inp = config.char_filter_size // 2 + config.lang_emb_size + config.word_emb_size + config.external_proj_size
        NUM_FILTERS = config.cnn_filter
        for _ in range(config.cnn_layers):
            conv_layer = nn.Sequential(
                ConvNorm(cs_inp,
                         NUM_FILTERS,
                         kernel_size=5, stride=1,
                         padding=2,
                         dilation=1, w_init_gain='tanh'),
                nn.BatchNorm1d(NUM_FILTERS))
            conv_layers.append(conv_layer)
            cs_inp = NUM_FILTERS // 2 + config.lang_emb_size

        ext2int = []
        for input_size in self._ext_word_emb:
            module = MLP(input_size, config.external_proj_size)
            ext2int.append(module)
        self._ext_proj = nn.ModuleList(ext2int)
        self._word_emb = nn.Embedding(len(encodings.word2int), config.word_emb_size, padding_idx=0)
        self._lang_emb = nn.Embedding(encodings.num_langs + 1, config.lang_emb_size, padding_idx=0)
        self._convs = nn.ModuleList(conv_layers)
        self._upos = LinearNorm(NUM_FILTERS // 2 + config.lang_emb_size, len(encodings.upos2int))
        self._upos_emb = nn.Embedding(len(encodings.upos2int), 64)
        self._xpos = LinearNorm(64 + config.lang_emb_size + NUM_FILTERS // 2, len(encodings.xpos2int))
        self._attrs = LinearNorm(64 + config.lang_emb_size + NUM_FILTERS // 2, len(encodings.attrs2int))

        self._aupos = LinearNorm(config.char_filter_size // 2 + config.lang_emb_size, len(encodings.upos2int))
        self._axpos = LinearNorm(config.char_filter_size // 2 + config.lang_emb_size, len(encodings.xpos2int))
        self._aattrs = LinearNorm(config.char_filter_size // 2 + config.lang_emb_size, len(encodings.attrs2int))

        if self._language_codes:
            self._res = {}
            for language_code in self._language_codes:
                self._res[language_code] = {"upos": 0., "xpos": 0., "attrs": 0.}
            self._early_stop_meta_val = 0

    def _compute_early_stop(self, res):
        for lang in res:
            if res[lang]["upos"] > self._res[lang]["upos"]:
                self._early_stop_meta_val += 1
                self._res[lang]["upos"] = res[lang]["upos"]
                res[lang]["upos_best"] = True
            if res[lang]["xpos"] > self._res[lang]["xpos"]:
                self._early_stop_meta_val += 1
                self._res[lang]["xpos"] = res[lang]["xpos"]
                res[lang]["xpos_best"] = True
            if res[lang]["attrs"] > self._res[lang]["attrs"]:
                self._early_stop_meta_val += 1
                self._res[lang]["attrs"] = res[lang]["attrs"]
                res[lang]["attrs_best"] = True
        return res

    def forward(self, X):
        x_sents = X['x_sent']
        x_lang_sent = X['x_lang_sent']
        x_words_chars = X['x_word']
        x_words_case = X['x_word_case']
        x_lang_word = X['x_lang_word']
        x_sent_len = X['x_sent_len']
        x_word_len = X['x_word_len']
        x_sent_masks = X['x_sent_masks']
        x_word_masks = X['x_word_masks']
        x_word_emb_packed = X['x_word_embeddings']
        char_emb_packed = self._word_net(x_words_chars, x_words_case, x_lang_word, x_word_masks, x_word_len)
        gs_upos = None
        if 'y_upos' in X:
            gs_upos = X['y_upos']
        sl = x_sent_len.cpu().numpy()

        char_emb = unpack(char_emb_packed, sl, x_sents.shape[1], device=self._get_device())
        word_emb_ext = None

        for ii in range(len(x_word_emb_packed)):
            we = unpack(x_word_emb_packed[ii], sl, x_sents.shape[1], self._get_device())
            if word_emb_ext is None:
                word_emb_ext = self._ext_proj[ii](we)
            else:
                word_emb_ext = word_emb_ext + self._ext_proj[ii](we)

        word_emb_ext = word_emb_ext / len(x_word_emb_packed)
        word_emb_ext = torch.tanh(word_emb_ext)

        lang_emb = self._lang_emb(x_lang_sent)
        lang_emb = lang_emb.unsqueeze(1).repeat(1, char_emb.shape[1], 1)

        aupos = self._aupos(torch.cat([char_emb, lang_emb], dim=-1))
        axpos = self._axpos(torch.cat([char_emb, lang_emb], dim=-1))
        aattrs = self._aattrs(torch.cat([char_emb, lang_emb], dim=-1))

        word_emb = self._word_emb(x_sents)

        x = mask_concat([word_emb, char_emb, word_emb_ext], 0.33, self.training, self._get_device())
        x = torch.cat([x, lang_emb], dim=-1)
        x = x.permute(0, 2, 1)
        lang_emb = lang_emb.permute(0, 2, 1)
        half = self._config.cnn_filter // 2
        res = None
        cnt = 0
        for conv in self._convs:
            conv_out = conv(x)
            tmp = torch.tanh(conv_out[:, :half, :]) * torch.sigmoid((conv_out[:, half:, :]))
            if res is None:
                res = tmp
            else:
                res = res + tmp
            x = torch.dropout(tmp, 0.2, self.training)
            cnt += 1
            if cnt != self._config.cnn_layers:
                x = torch.cat([x, lang_emb], dim=1)
        x = x + res
        xu = x.permute(0, 2, 1).contiguous()
        x = torch.cat([x, lang_emb], dim=1)
        x = x.permute(0, 2, 1)
        # x = torch.tanh(x)
        upos = self._upos(x)
        if gs_upos is None:
            upos_idx = torch.argmax(upos, dim=-1)
        else:
            upos_idx = gs_upos
        upos_emb = self._upos_emb(upos_idx)
        upos_emb = torch.cat([upos_emb, lang_emb.permute(0, 2, 1)], dim=-1)
        xpos = self._xpos(torch.cat([upos_emb, xu], dim=-1))
        attrs = self._attrs(torch.cat([upos_emb, xu], dim=-1))
        return upos, xpos, attrs, aupos, axpos, aattrs

    def _get_device(self):
        if self._lang_emb.weight.device.type == 'cpu':
            return 'cpu'
        return '{0}:{1}'.format(self._lang_emb.weight.device.type, str(self._lang_emb.weight.device.index))

    def configure_optimizers(self):
        optimizer = torch.optim.AdamW(self.parameters())
        return optimizer

    def training_step(self, batch, batch_idx):
        p_upos, p_xpos, p_attrs, a_upos, a_xpos, a_attrs = self.forward(batch)
        y_upos = batch['y_upos']
        y_xpos = batch['y_xpos']
        y_attrs = batch['y_attrs']

        loss_upos = F.cross_entropy(p_upos.view(-1, p_upos.shape[2]), y_upos.view(-1), ignore_index=0)
        loss_xpos = F.cross_entropy(p_xpos.reshape(-1, p_xpos.shape[2]), y_xpos.view(-1), ignore_index=0)
        loss_attrs = F.cross_entropy(p_attrs.reshape(-1, p_attrs.shape[2]), y_attrs.view(-1), ignore_index=0)

        loss_aupos = F.cross_entropy(a_upos.view(-1, a_upos.shape[2]), y_upos.view(-1), ignore_index=0)
        loss_axpos = F.cross_entropy(a_xpos.view(-1, a_xpos.shape[2]), y_xpos.view(-1), ignore_index=0)
        loss_aattrs = F.cross_entropy(a_attrs.view(-1, a_attrs.shape[2]), y_attrs.view(-1), ignore_index=0)

        step_loss = ((loss_upos + loss_attrs + loss_xpos) / 3.) * 1.0 + (
                (loss_aupos + loss_aattrs + loss_axpos) / 3.) * 1.0

        return {'loss': step_loss}

    def validation_step(self, batch, batch_idx):
        y_upos = batch['y_upos']
        y_xpos = batch['y_xpos']
        y_attrs = batch['y_attrs']
        x_sent_len = batch['x_sent_len']
        x_lang = batch['x_lang_sent']
        del batch['y_upos']
        p_upos, p_xpos, p_attrs, a_upos, a_xpos, a_attrs = self.forward(batch)

        loss_upos = F.cross_entropy(p_upos.view(-1, p_upos.shape[2]), y_upos.view(-1), ignore_index=0)
        loss_xpos = F.cross_entropy(p_xpos.reshape(-1, p_xpos.shape[2]), y_xpos.view(-1), ignore_index=0)
        loss_attrs = F.cross_entropy(p_attrs.reshape(-1, p_attrs.shape[2]), y_attrs.view(-1), ignore_index=0)
        loss = (loss_upos + loss_attrs + loss_xpos) / 3
        language_result = {lang_id: {'total': 0, 'upos_ok': 0, 'xpos_ok': 0, 'attrs_ok': 0} for lang_id in
                           range(self._num_langs)}

        pred_upos = torch.argmax(p_upos, dim=-1).detach().cpu().numpy()
        pred_xpos = torch.argmax(p_xpos, dim=-1).detach().cpu().numpy()
        pred_attrs = torch.argmax(p_attrs, dim=-1).detach().cpu().numpy()
        tar_upos = y_upos.detach().cpu().numpy()
        tar_xpos = y_xpos.detach().cpu().numpy()
        tar_attrs = y_attrs.detach().cpu().numpy()
        sl = x_sent_len.detach().cpu().numpy()
        x_lang = x_lang.detach().cpu().numpy()
        for iSent in range(p_upos.shape[0]):
            for iWord in range(sl[iSent]):
                lang_id = x_lang[iSent] - 1
                language_result[lang_id]['total'] += 1
                if pred_upos[iSent, iWord] == tar_upos[iSent, iWord]:
                    language_result[lang_id]['upos_ok'] += 1
                if pred_xpos[iSent, iWord] == tar_xpos[iSent, iWord]:
                    language_result[lang_id]['xpos_ok'] += 1
                if pred_attrs[iSent, iWord] == tar_attrs[iSent, iWord]:
                    language_result[lang_id]['attrs_ok'] += 1

        return {'loss': loss, 'acc': language_result}

    def validation_epoch_end(self, outputs):
        language_result = {lang_index: {'total': 0, 'upos_ok': 0, 'xpos_ok': 0, 'attrs_ok': 0} for lang_index in
                           range(self._num_langs)}

        valid_loss_total = 0
        total = 0
        attrs_ok = 0
        upos_ok = 0
        xpos_ok = 0
        for out in outputs:
            valid_loss_total += out['loss']
            for lang_index in language_result:
                valid_loss_total += out['loss']
                language_result[lang_index]['total'] += out['acc'][lang_index]['total']
                language_result[lang_index]['upos_ok'] += out['acc'][lang_index]['upos_ok']
                language_result[lang_index]['xpos_ok'] += out['acc'][lang_index]['xpos_ok']
                language_result[lang_index]['attrs_ok'] += out['acc'][lang_index]['attrs_ok']
                # global
                total += out['acc'][lang_index]['total']
                upos_ok += out['acc'][lang_index]['upos_ok']
                xpos_ok += out['acc'][lang_index]['xpos_ok']
                attrs_ok += out['acc'][lang_index]['attrs_ok']

        self.log('val/loss', valid_loss_total / len(outputs))
        self.log('val/UPOS/total', upos_ok / total)
        self.log('val/XPOS/total', xpos_ok / total)
        self.log('val/ATTRS/total', attrs_ok / total)

        res = {}
        for lang_index in language_result:
            total = language_result[lang_index]['total']
            if total == 0:
                total = 1
            if self._language_codes is None:
                lang = lang_index
            else:
                lang = self._language_codes[lang_index]
            res[lang] = {
                "upos": language_result[lang_index]['upos_ok'] / total,
                "xpos": language_result[lang_index]['xpos_ok'] / total,
                "attrs": language_result[lang_index]['attrs_ok'] / total
            }

            self.log('val/UPOS/{0}'.format(lang), language_result[lang_index]['upos_ok'] / total)
            self.log('val/XPOS/{0}'.format(lang), language_result[lang_index]['xpos_ok'] / total)
            self.log('val/ATTRS/{0}'.format(lang), language_result[lang_index]['attrs_ok'] / total)

        # single value for early stopping
        self._epoch_results = self._compute_early_stop(res)
        self.log('val/early_meta', self._early_stop_meta_val)

        # print("\n\n\n", upos_ok / total, xpos_ok / total, attrs_ok / total,
        #      aupos_ok / total, axpos_ok / total, aattrs_ok / total, "\n\n\n")

    def load(self, model_path:str, device: str = 'cpu'):
        self.load_state_dict(torch.load(model_path, map_location='cpu')['state_dict'])
        self.to(device)

    def process(self, doc: Document, collate: MorphoCollate, upos: bool = True, xpos: bool = True, attrs: bool = True,
                batch_size: int = 32, num_workers: int = 4) -> Document:
        self.eval()
        if not (upos or xpos or attrs):
            raise Exception("To perform tagging at least one of 'upos', 'xpos' or 'attrs' must be set to True.")

        dataset = MorphoDataset(doc)

        dataloader = DataLoader(dataset, batch_size=batch_size, collate_fn=collate.collate_fn,
                                shuffle=False, num_workers=num_workers, pin_memory=True)
        index = 0

        with torch.no_grad():
            for batch in dataloader:
                del batch['y_upos']
                p_upos, p_xpos, p_attrs, _, _, _ = self.forward(batch)

                batch_size = p_upos.size()[0]

                if upos:
                    pred_upos = torch.argmax(p_upos, dim=-1).detach().cpu().numpy()
                if xpos:
                    pred_xpos = torch.argmax(p_xpos, dim=-1).detach().cpu().numpy()
                if attrs:
                    pred_attrs = torch.argmax(p_attrs, dim=-1).detach().cpu().numpy()

                for sentence_index in range(batch_size):  # for each sentence
                    # print(f"at index {index+sentence_index}, sentence {sentence_index} has {batch['x_sent_len'][sentence_index]} words.")
                    for word_index in range(batch["x_sent_len"][sentence_index]):
                        if upos:
                            doc.sentences[index + sentence_index].words[word_index].upos = self._encodings.upos_list[
                                pred_upos[sentence_index][word_index]]
                        if xpos:
                            doc.sentences[index + sentence_index].words[word_index].xpos = self._encodings.xpos_list[
                                pred_xpos[sentence_index][word_index]]
                        if attrs:
                            doc.sentences[index + sentence_index].words[word_index].attrs = self._encodings.attrs_list[
                                pred_attrs[sentence_index][word_index]]

                index += batch_size
        #                break
        return doc

    class PrintAndSaveCallback(pl.callbacks.Callback):
        def __init__(self, store_prefix):
            super().__init__()
            self.store_prefix = store_prefix

        def on_validation_end(self, trainer, pl_module):
            metrics = trainer.callback_metrics
            epoch = trainer.current_epoch

            for lang in pl_module._epoch_results:
                res = pl_module._epoch_results[lang]
                if "upos_best" in res:
                    trainer.save_checkpoint(self.store_prefix + "." + lang + ".upos")
                if "xpos_best" in res:
                    trainer.save_checkpoint(self.store_prefix + "." + lang + ".xpos")
                if "attrs_best" in res:
                    trainer.save_checkpoint(self.store_prefix + "." + lang + ".attrs")

            trainer.save_checkpoint(self.store_prefix + ".last")

            s = "{0:30s}\tUPOS\tXPOS\tATTRS".format("Language")
            print("\n\n\t" + s)
            print("\t" + ("=" * (len(s) + 9)))
            for lang in pl_module._language_codes:
                upos = metrics["val/UPOS/{0}".format(lang)]
                xpos = metrics["val/XPOS/{0}".format(lang)]
                attrs = metrics["val/ATTRS/{0}".format(lang)]
                msg = "\t{0:30s}:\t{1:.4f}\t{2:.4f}\t{3:.4f}".format(lang, upos, xpos, attrs)
                print(msg)
            print("\n")


if __name__ == '__main__':

    root = "data/be"
    language_code = "be_hse"
    device = 'cpu'  # 'cuda'
    batch_size = 2

    # read yaml
    object_config = yaml.full_load(open(root + ".yaml"))

    # read model config
    config = TaggerConfig(filename=root + ".config")

    # read encodings
    encodings = Encodings()
    encodings.load(filename=root + ".encodings")

    # load models
    tagger_UPOS = Tagger.load_from_checkpoint(root + "." + language_code + ".upos", config=config, encodings=encodings,
                                              language_codes=object_config["language_codes"])
    tagger_UPOS.to(device)
    tagger_UPOS.eval()
    tagger_UPOS.freeze()

    tagger_XPOS = Tagger.load_from_checkpoint(root + "." + language_code + ".xpos", config=config, encodings=encodings,
                                              language_codes=object_config["language_codes"])
    tagger_XPOS.to(device)
    tagger_XPOS.eval()
    tagger_XPOS.freeze()

    tagger_ATTRS = Tagger.load_from_checkpoint(root + "." + language_code + ".attrs", config=config,
                                               encodings=encodings,
                                               language_codes=object_config["language_codes"])
    tagger_ATTRS.to(device)
    tagger_ATTRS.eval()
    tagger_ATTRS.freeze()

    # read a doc
    doc = Document()
    doc.load("corpus/ud-treebanks-v2.5/UD_Belarusian-HSE/be_hse-ud-dev.conllu",
             lang_id=object_config["language_codes"].index(language_code))

    for si, _ in enumerate(doc.sentences):
        for wi, _ in enumerate(doc.sentences[si].words):
            doc.sentences[si].words[wi].upos = "<change>"
            doc.sentences[si].words[wi].xpos = "<change>"
            doc.sentences[si].words[wi].attrs = "<change>"

    print(doc)

    doc = tagger_UPOS.process(doc, upos=True, xpos=False, attrs=False, batch_size=batch_size)
    doc = tagger_XPOS.process(doc, upos=False, xpos=True, attrs=False, batch_size=batch_size)
    doc = tagger_ATTRS.process(doc, upos=False, xpos=False, attrs=True, batch_size=batch_size)

    print(doc)



import sys
from typing import *

import torch
import torch.nn as nn
import torch.nn.functional as F
import pytorch_lightning as pl
import numpy as np
from torch.utils.data import DataLoader

from cube.io_utils.objects import Document
from cube.networks.utils import LemmaDataset, Word2TargetCollate

sys.path.append('')
from cube.io_utils.encodings import Encodings
from cube.io_utils.config import LemmatizerConfig
from cube.networks.modules import LinearNorm, ConvNorm, Attention


class Lemmatizer(pl.LightningModule):
    encodings: Encodings
    config: LemmatizerConfig

    def __init__(self, config: LemmatizerConfig, encodings: Encodings, language_codes: [] = None):
        super(Lemmatizer, self).__init__()
        NUM_FILTERS = 512
        NUM_LAYERS = 5
        self._config = config
        self._encodings = encodings
        self._num_languages = encodings.num_langs
        self._language_codes = language_codes
        self._eol = len(encodings.char2int)
        self._num_filters = NUM_FILTERS

        self._char_list = ['' for char in encodings.char2int]
        for char in encodings.char2int:
            self._char_list[encodings.char2int[char]] = char
        self._lang_emb = nn.Embedding(self._num_languages + 1, config.lang_emb_size, padding_idx=0)
        self._upos_emb = nn.Embedding(len(encodings.upos2int), config.upos_emb_size, padding_idx=0)
        self._char_emb = nn.Embedding(len(encodings.char2int) + 2, config.char_emb_size,
                                      padding_idx=0)  # start/stop index
        self._case_emb = nn.Embedding(4, 16, padding_idx=0)  # 0-pad 1-symbol 2-upper 3-lower
        convolutions = []
        cs_inp = config.char_emb_size + config.lang_emb_size + config.upos_emb_size + 16

        for _ in range(NUM_LAYERS):
            conv_layer = nn.Sequential(
                ConvNorm(cs_inp,
                         NUM_FILTERS,
                         kernel_size=5, stride=1,
                         padding=2,
                         dilation=1, w_init_gain='tanh'),
                nn.BatchNorm1d(NUM_FILTERS))
            convolutions.append(conv_layer)
            cs_inp = NUM_FILTERS // 2 + config.lang_emb_size

        self._convolutions_char = nn.ModuleList(convolutions)
        self._decoder = nn.LSTM(
            NUM_FILTERS // 2 + config.char_emb_size + config.lang_emb_size + config.upos_emb_size + 16,
            config.decoder_size, config.decoder_layers,
            batch_first=True, bidirectional=False)
        self._attention = Attention(
            (NUM_FILTERS // 2 + config.lang_emb_size + config.upos_emb_size + 16) // 2,
            config.decoder_size, config.att_proj_size)

        self._output_char = LinearNorm(config.decoder_size, len(self._encodings.char2int) + 2)
        self._output_case = LinearNorm(config.decoder_size, 4)
        self._start_frame = nn.Embedding(1,
                                         NUM_FILTERS // 2 + config.char_emb_size + config.lang_emb_size + config.upos_emb_size + 16)

        if self._language_codes:
            self._res = {}
            for language_code in self._language_codes:
                self._res[language_code] = {"loss": 0., "acc": 0.}
            self._early_stop_meta_val = 0
            self._epoch_results = None

    def forward(self, X):
        x_char = X['x_char']
        x_case = X['x_case']
        x_lang = X['x_lang']
        x_upos = X['x_upos']

        if 'y_char' in X:
            gs_output = X['y_char']
        else:
            gs_output = None

        char_emb = self._char_emb(x_char)
        case_emb = self._case_emb(x_case)

        upos_emb = self._upos_emb(x_upos).unsqueeze(1).repeat(1, char_emb.shape[1], 1)
        lang_emb = self._lang_emb(x_lang).unsqueeze(1).repeat(1, char_emb.shape[1], 1)
        conditioning = torch.cat((case_emb, upos_emb), dim=-1)
        if gs_output is not None:
            output_idx = gs_output

        x = torch.cat((char_emb, conditioning), dim=-1)
        half = self._num_filters // 2
        count = 0
        res = None
        skip = None
        x_lang_conv = lang_emb.permute(0, 2, 1)
        x = x.permute(0, 2, 1)
        for conv in self._convolutions_char:
            count += 1
            drop = self.training
            if count >= len(self._convolutions_char):
                drop = False
            if skip is not None:
                x = x + skip

            x = torch.cat([x, x_lang_conv], dim=1)
            conv_out = conv(x)
            tmp = torch.tanh(conv_out[:, :half, :]) * torch.sigmoid((conv_out[:, half:, :]))
            if res is None:
                res = tmp
            else:
                res = res + tmp
            skip = tmp
            x = torch.dropout(tmp, 0.1, drop)
        x = x + res
        x = x.permute(0, 2, 1)
        encoder_output = torch.cat((x, conditioning, lang_emb), dim=-1)

        step = 0
        done = np.zeros(encoder_output.shape[0])
        start_frame = self._start_frame(
            torch.tensor([0], dtype=torch.long, device=self._get_device())).unsqueeze(1).repeat(encoder_output.shape[0],
                                                                                                1, 1)
        decoder_output, decoder_hidden = self._decoder(start_frame)

        out_char_list = []
        out_case_list = []
        while True:
            if gs_output is not None:
                if step == output_idx.shape[1]:
                    break
            elif np.sum(done) == encoder_output.shape[0]:
                break
            elif step == encoder_output.shape[1] * 20:  # failsafe
                break

            att = self._attention(decoder_hidden[-1][-1, :, :], encoder_output)
            context = torch.bmm(att.unsqueeze(1), encoder_output)

            if step == 0:
                prev_char_emb = torch.zeros((encoder_output.shape[0], 1, self._config.char_emb_size),
                                            device=self._get_device())

            decoder_input = torch.cat((context, prev_char_emb), dim=-1)
            decoder_output, decoder_hidden = self._decoder(decoder_input,
                                                           hx=(torch.dropout(decoder_hidden[0], 0.5, self.training),
                                                               torch.dropout(decoder_hidden[1], 0.5, self.training)))

            output_char = self._output_char(decoder_output)
            output_case = self._output_case(decoder_output)
            out_char_list.append(output_char)
            out_case_list.append(output_case)
            selected_chars = torch.argmax(output_char, dim=-1)
            for ii in range(selected_chars.shape[0]):
                if selected_chars[ii].squeeze() == self._eol:
                    done[ii] = 1
            if gs_output is not None:
                prev_char_emb = self._char_emb(output_idx[:, step]).unsqueeze(1)
            else:
                prev_char_emb = self._char_emb(selected_chars)

            step += 1

        return torch.cat(out_char_list, dim=1), torch.cat(out_case_list, dim=1)

    def save(self, path):
        torch.save(self.state_dict(), path)

    def load(self, model_path: str, device: str = 'cpu'):
        self.load_state_dict(torch.load(model_path, map_location='cpu')['state_dict'])
        self.to(device)

    def _get_device(self):
        if self._char_emb.weight.device.type == 'cpu':
            return 'cpu'
        return '{0}:{1}'.format(self._char_emb.weight.device.type, str(self._char_emb.weight.device.index))

    def process(self, doc: Document, collate: Word2TargetCollate, batch_size: int = 4,
                num_workers: int = 4) -> Document:
        self.eval()
        dataset = LemmaDataset(doc, for_training=False)

        dataloader = DataLoader(dataset, batch_size=batch_size, collate_fn=collate.collate_fn,
                                shuffle=False, num_workers=num_workers)

        data_iterator = iter(dataloader)

        end_char_value = len(self._encodings.char2int)

        with torch.no_grad():
            all_lemmas = []
            for batch in dataloader:
                del batch['y_char']  # set for prediction, not training
                del batch['y_case']

                for key in batch:
                    if isinstance(batch[key], torch.Tensor):
                        batch[key] = batch[key].to(self._device)

                y_char_pred, y_case_pred = self.forward(batch)
                y_char_pred = torch.argmax(y_char_pred.detach(), dim=-1).cpu().numpy()  # list of lists of int
                y_case_pred = torch.argmax(y_case_pred.detach(), dim=-1).cpu().numpy()  # list of lists of int
                for word_index in range(y_char_pred.shape[0]):
                    # get letters
                    lemma = []
                    for char_val, case_val in zip(y_char_pred[word_index],
                                                  y_case_pred[word_index]):  # [[24, 12, 88]], get the inside list
                        if char_val == end_char_value:
                            break
                        chr = self._encodings.characters[char_val]
                        if case_val == 2:
                            chr = chr.upper()
                        elif case_val == 3:
                            chr = chr.lower()
                        lemma.append(chr)

                    all_lemmas.append("".join(lemma))
            lemma_index = 0
            for sentence_index in range(len(doc.sentences)):
                for word_index in range(len(doc.sentences[sentence_index].words)):
                    doc.sentences[sentence_index].words[word_index].lemma = all_lemmas[lemma_index]
                    lemma_index += 1

        return doc

    def configure_optimizers(self):
        return torch.optim.AdamW(self.parameters())

    def training_step(self, batch, batch_idx):
        y_char_pred, y_case_pred = self.forward(batch)
        y_char_target, y_case_target = batch['y_char'], batch['y_case']
        loss_char = F.cross_entropy(y_char_pred.view(-1, y_char_pred.shape[2]), y_char_target.view(-1), ignore_index=0)
        loss_case = F.cross_entropy(y_case_pred.view(-1, y_case_pred.shape[2]), y_case_target.view(-1), ignore_index=0)
        return loss_char + loss_case

    def validation_step(self, batch, batch_idx):
        y_char_target, y_case_target = batch['y_char'], batch['y_case']
        del batch['y_char']
        y_char_pred, y_case_pred = self.forward(batch)
        language_result = {lang_id: {'total': 0, 'ok': 0}
                           for lang_id in range(self._num_languages)}

        y_char_target = y_char_target.detach().cpu().numpy()
        y_char_pred = torch.argmax(y_char_pred.detach(), dim=-1).cpu().numpy()
        lang = batch['x_lang'].detach().cpu().numpy()
        for lang_id, y_pred, y_target in zip(lang, y_char_pred, y_char_target):
            valid = True
            for y_p, y_t in zip(y_pred, y_target):
                if y_t != 0 and y_p != y_t:
                    valid = False
                    break
            if valid:
                language_result[lang_id - 1]['ok'] += 1
            language_result[lang_id - 1]['total'] += 1

        return {'acc': language_result}

    def validation_epoch_end(self, outputs: List[Any]) -> None:
        language_result = {lang_id: {'total': 0, 'ok': 0}
                           for lang_id in
                           range(self._num_languages)}
        for result in outputs:
            for lang_id in result['acc']:
                language_result[lang_id]['ok'] += result['acc'][lang_id]['ok']
                language_result[lang_id]['total'] += result['acc'][lang_id]['total']

        res = {}
        for lang_index in language_result:
            total = language_result[lang_index]['total']
            if total == 0:
                total = 1
            if self._language_codes is None:
                lang = lang_index
            else:
                lang = self._language_codes[lang_index]
            res[lang] = {
                "acc": language_result[lang_index]['ok'] / total,
            }

            self.log('val/ACC/{0}'.format(lang), language_result[lang_index]['ok'] / total)

        # single value for early stopping
        self._epoch_results = self._compute_early_stop(res)
        self.log('val/early_meta', self._early_stop_meta_val)

    def _compute_early_stop(self, res):
        for lang in res:
            if res[lang]["acc"] > self._res[lang]["acc"]:
                self._early_stop_meta_val += 1
                self._res[lang]["acc"] = res[lang]["acc"]
                res[lang]["acc_best"] = True
        return res

    class PrintAndSaveCallback(pl.callbacks.Callback):
        def __init__(self, store_prefix):
            super().__init__()
            self.store_prefix = store_prefix

        def on_validation_end(self, trainer, pl_module):
            metrics = trainer.callback_metrics
            epoch = trainer.current_epoch

            for lang in pl_module._epoch_results:
                res = pl_module._epoch_results[lang]
                if "acc_best" in res:
                    trainer.save_checkpoint(self.store_prefix + "." + lang + ".best")

            trainer.save_checkpoint(self.store_prefix + ".last")

            s = "{0:30s}\tACC".format("Language")
            print("\n\n\t" + s)
            print("\t" + ("=" * (len(s) + 16)))
            for lang in pl_module._language_codes:
                acc = metrics["val/ACC/{0}".format(lang)]
                msg = "\t{0:30s}:\t{1:.4f}".format(lang, acc)
                print(msg)
            print("\n")
